# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

""" vnx.py: Module to handle commands on EMC VNX Array.
    We assume  NaviSphere (naviseccli) is installed """

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re  # regex

import libsan.host.fc
import libsan.host.iscsi
import libsan.host.linux
import libsan.misc.array
import libsan.misc.size
from libsan import _print
from libsan.host.cmdline import run


def _size_human_2_size_netapp(size_human):
    if not size_human:
        _print("FAIL: _size_human_2_size_netapp() - requires size_human parameter")
        return None

    if not libsan.misc.size.size_human_check(size_human):
        _print("FAIL: _size_human_2_size_netapp() - Checking input")
        return None

    m = libsan.misc.size.size_human_regex.match(size_human)
    if m.group(2):
        unit = m.group(2)
        unit = unit.replace("i", "")
        unit = unit.lower()
        return f"{m.group(1)}{unit}"
    return None


def _lun_uid2wwid(uid):
    """Usage
        _lun_uid2wwid(uid)
    Purpose
        Convert VNC UID to WWID.
        1. It simply add 3 at the front
        2. remove all :
    Parameter
        uid:        # like 60:06:01:60:A3:26:11:00:DD:C5:51:79:9F:F8:E1:11
    Returns
        wwid:       # like '360060160a3261100008c2782b292e211'
    """
    if not uid:
        return None

    wwid = "3%s" % uid

    wwid = wwid.replace(":", "")
    wwid = wwid.lower()
    return wwid


def _uid2wwnn_wwpn(uid):
    """Convert UID format to WWNN and WWPN
    Return a dict
    """
    if not uid:
        _print("FAIL: _uid2wwnn_wwpn() - requires uid as parameter")
        return None
    uid = libsan.host.fc.standardize_wwpn(uid)
    if not uid:
        # uid for example, can be an iqn
        # _print("FAIL: _uid2wwnn_wwpn() - Could not standardize uid")
        return None

    emc_wwpn_regex = "(?:[0-9a-f]{2}:){7}[0-9a-f]{2}"
    emc_wwnn_regex = emc_wwpn_regex
    emc_hba_id = re.compile(f"({emc_wwnn_regex}):({emc_wwpn_regex})")

    m = emc_hba_id.match(uid)
    if not m:
        return None

    ret = {}
    ret["wwnn"] = libsan.host.fc.standardize_wwpn(m.group(1))
    ret["wwpn"] = libsan.host.fc.standardize_wwpn(m.group(2))
    return ret


class vnx:  # noqa: N801
    """Class to manage VNX array"""

    host = None
    user = None
    passwd = None

    navi_cli_path = "/opt/Navisphere/bin/naviseccli"
    navi_scope = 0
    ctrl_storage_pool = None

    san_conf_path = None
    sa_conf_dict: dict = {}

    luns_dict: dict = {}
    storage_groups_dict: dict = {}
    ssports: list = []

    map_info: list = []
    BLOCK_SIZE_EMC_VNX_VNX = 512

    # Used for getall command
    get_all_output = None

    def __init__(self, hostname, username, password, timeout=None, san_dev=None):
        # san_dev is here because we need it for DELL Equalogic array class
        self.san_dev_name = san_dev
        self.host = hostname
        self.user = username
        self.passwd = password
        self.timeout = timeout

    def set_san_conf_path(self, san_conf_path):
        self.san_conf_path = san_conf_path
        return True

    def set_sa_conf(self, sa_conf_dict):
        self.sa_conf_dict = sa_conf_dict
        if self.sa_conf_dict:
            if "navi_cli_path" in list(self.sa_conf_dict.keys()):
                self.navi_cli_path = self.sa_conf_dict["navi_cli_path"]

            if "navi_scope" in list(self.sa_conf_dict.keys()):
                self.navi_scope = self.sa_conf_dict["navi_scope"]

            if "navi_scope" in list(self.sa_conf_dict.keys()):
                self.navi_scope = self.sa_conf_dict["navi_scope"]

            if "ctrl_storage_pool" in list(self.sa_conf_dict.keys()):
                self.ctrl_storage_pool = self.sa_conf_dict["ctrl_storage_pool"]

        return True

    def get_sa_conf(self):
        return self.sa_conf_dict

    def _run(self, cmd, ctrler=None, return_output=False, verbose=True):
        if not ctrler:
            ctrler = self.host

        _print("INFO: Connecting to %s for command:" % ctrler)

        navi_prefix = "{} -User {} -Scope {} -h {} -Password {}".format(
            self.navi_cli_path,
            self.user,
            self.navi_scope,
            ctrler,
            self.passwd,
        )
        if self.timeout:
            navi_prefix += " -t %s" % self.timeout

        print(cmd)
        cmd = f"{navi_prefix} {cmd}"
        # If return_output is True, ret will be a tuple
        ret = run(cmd, return_output=return_output, verbose=verbose)
        # ret = ssh.run_cmd(session, cmd, return_output = return_output, verbose = verbose,
        #                  invoke_shell = False, timeout=self.timeout)
        # ssh.disconnect(session)
        return ret

    @staticmethod
    def get_version():
        """Return Firmware version running on array"""
        return

    @staticmethod
    def capability():
        """Indicates supported operation on array"""
        cap_dict = {}
        cap_dict["lun_info"] = True
        cap_dict["lun_query"] = True
        cap_dict["lun_create"] = True
        cap_dict["lun_map"] = True
        cap_dict["lun_unmap"] = True
        cap_dict["lun_remove"] = True
        cap_dict["lun_grow"] = False
        cap_dict["lun_shrink"] = False
        cap_dict["lun_trepass"] = False
        cap_dict["lun_thinp"] = True
        cap_dict["sa_ctrler_reboot"] = True
        return cap_dict

    def _get_all(self, recheck=False):
        """ """
        if self.get_all_output and not recheck:
            return self.get_all_output

        cmd = "getall"
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: "%s" command' % cmd)
            print(output)
            return None

        if not output:
            return None

        self.get_all_output = None
        self.get_all_output = output.split("\n")
        return self.get_all_output

    def query_all_luns(self, recheck=False):
        """ """
        all_luns_info = self.query_all_lun_info(recheck=recheck)
        if not all_luns_info:
            return None
        return sorted(all_luns_info.keys())

    def query_all_lun_info(self, recheck=False):
        """Query all LUNs on array and store its information on a dict"""

        # If we do not need to recheck and port info exist we return it
        # if self.luns_dict and not recheck:
        #    return self.luns_dict

        self.luns_dict = {}

        # If not recheck, it will return exiting information
        getall_info = self._get_all(recheck=recheck)
        if not getall_info:
            return None

        storage_grp_info = self.query_all_storage_group_info()
        # _print("DEBUG: query_all_lun_info() - dump storage grps:")
        # print storage_grp_info

        _lun_section_regex = re.compile(r"All logical Units Information")
        _raid_section_regex = re.compile(r"All RAID Groups Information")
        _lun_id_regex = re.compile(r"LOGICAL UNIT NUMBER (\d+)$")
        # If no name was given to the LUN they will be called like "LUN 10"
        # so they might contain space characters
        _lun_name_regex = re.compile(r"Name\s+(\S+.*)$")
        _lun_capacity_regex = re.compile(r"LUN Capacity\(Blocks\):\s+(\d+)$")
        _lun_default_owner_regex = re.compile(r"Default [Oo]wner: +SP ([A-Z]+)")
        _lun_current_owner_regex = re.compile(r"Current [Oo]wner: +SP ([A-Z]+)")
        _lun_uid_regex = re.compile(r"UID:[ \t]+([0-9A-F:]+)$")

        flag_all_lun_info_begin = False
        lun_info = None

        # Lun ID is the first item that is listed for a LUN
        # We store it to the lun info once we get the LUN name
        lun_id = None

        for info in getall_info:
            if _lun_section_regex.match(info):
                flag_all_lun_info_begin = True
                continue
            if _raid_section_regex.match(info):
                flag_all_lun_info_begin = False
                continue

            if not flag_all_lun_info_begin:
                continue

            # _print("DEBUG: LUN info '%s'" % info)
            m = _lun_id_regex.match(info)
            if m:
                lun_id = m.group(1)
                continue

            m = _lun_name_regex.match(info)
            if m:
                if m.group(1) in list(self.luns_dict.keys()):
                    _print("FAIL: query_all_lun_info() - Found lun name (%s) twice" % m.group(1))
                    print(getall_info)
                    continue
                self.luns_dict[m.group(1)] = {}
                lun_info = self.luns_dict[m.group(1)]
                lun_info["name"] = m.group(1)
                lun_info["lun_id"] = lun_id
                continue

            # Skip while LUN name is not found
            if not lun_info:
                continue

            m = _lun_capacity_regex.match(info)
            if m:
                if not lun_info:
                    _print("FAIL: Does not know to which lun add capacity: %s" % m.group(1))
                    continue
                if "size" in list(lun_info.keys()) or "size_human" in list(lun_info.keys()):
                    _print(
                        "FAIL: query_all_lun_info() - It looks like we are parsing LUN info for LUN: %s twice"
                        % lun_info["name"]
                    )
                    return None
                lun_info["size"] = int(m.group(1)) * self.BLOCK_SIZE_EMC_VNX_VNX
                lun_info["size_human"] = libsan.misc.size.size_bytes_2_size_human(lun_info["size"])
                continue

            m = _lun_default_owner_regex.match(info)
            if m:
                if not lun_info:
                    _print("FAIL: Does not know to which lun add default owner: %s" % m.group(1))
                    continue
                lun_info["default_sp"] = m.group(1)

            m = _lun_current_owner_regex.match(info)
            if m:
                if not lun_info:
                    _print("FAIL: Does not know to which lun add current owner: %s" % m.group(1))
                    continue
                lun_info["current_sp"] = m.group(1)
                continue
            m = _lun_uid_regex.match(info)
            if m:
                if not lun_info:
                    _print("FAIL: Does not know to which lun add wwid: %s" % m.group(1))
                    continue
                lun_info["wwid"] = _lun_uid2wwid(m.group(1))

        # add map info
        # _print("DEBUG: Query all info - updating map info")
        # print storage_grp_info["st61_intel x520 boot"]
        for lun_info in list(self.luns_dict.values()):
            lun_info["map_infos"] = []

            if not storage_grp_info:
                continue

            for key in list(storage_grp_info.keys()):
                grp_info = storage_grp_info[key]
                if "map_infos" not in list(grp_info.keys()):
                    continue

                for grp_map_info in grp_info["map_infos"]:
                    if grp_map_info["t_lun_id"] == lun_info["lun_id"]:
                        # _print("DEBUG: processing map for st group %s for lun %s" % (key, lun_info["name"]))
                        # Check for intiiator information
                        if "hba_infos" in list(grp_info.keys()):
                            for hba_info in grp_info["hba_infos"]:
                                # _print ("DEBUG: %s" % hba_info)
                                map_info = {}
                                map_info["st_group_name"] = key
                                # map_info["t_lun_id"] = grp_map_info["t_lun_id"]
                                map_info["h_lun_id"] = grp_map_info["h_lun_id"]
                                map_info["sp_name"] = hba_info["sp_name"]
                                map_info["sp_port"] = hba_info["sp_port"]
                                if "t_wwpn" in list(hba_info.keys()):
                                    map_info["t_wwpn"] = hba_info["t_wwpn"]
                                if "h_wwpn" in list(hba_info.keys()):
                                    map_info["h_wwpn"] = hba_info["h_wwpn"]

                                if "t_iqn" in list(hba_info.keys()):
                                    map_info["t_iqn"] = hba_info["t_iqn"]
                                if "h_iqn" in list(hba_info.keys()):
                                    map_info["h_iqn"] = hba_info["h_iqn"]

                                # if map_info not in lun_info["map_infos"]:
                                lun_info["map_infos"].append(map_info)
                        else:
                            # the mapping can be done with only host and target lun IDs
                            if "h_lun_id" in list(grp_map_info.keys()):
                                map_info = {}
                                map_info["st_group_name"] = key
                                map_info["t_lun_id"] = grp_map_info["t_lun_id"]
                                map_info["h_lun_id"] = grp_map_info["h_lun_id"]
                                lun_info["map_infos"].append(map_info)

        return self.luns_dict

    def query_all_spport_info(self, recheck=False):
        """Query all SPPORt on array and store its information on a dict"""

        # If we do not need to recheck and port info exist we return it
        # if self.ssports and not recheck:
        #    return self.ssports

        self.ssports = []

        # If not recheck, it will return exiting information
        getall_info = self._get_all(recheck=recheck)
        if not getall_info:
            return None

        _spport_section_regex = re.compile(r"Information about each SPPORT:")
        _sp_info_section_regex = re.compile(r"^SP Information:$")
        _spport_name_regex = re.compile(r"SP Name:\s+(SP \S+)$")
        _spport_port_id_regex = re.compile(r"SP Port ID:\s+(\d+)$")
        _spport_port_uid_regex = re.compile(r"SP UID:\s+(\S+)$")

        flag_all_spport_info_begin = False
        ssport_info = None

        # Lun ID is the first item that is listed for a LUN
        # We store it to the lun info once we get the LUN name

        for info in getall_info:
            if _spport_section_regex.match(info):
                flag_all_spport_info_begin = True
                continue
            if _sp_info_section_regex.match(info):
                flag_all_spport_info_begin = False
                continue

            if not flag_all_spport_info_begin:
                continue

            m = _spport_name_regex.match(info)
            if m:
                ssport_info = {}
                self.ssports.append(ssport_info)
                ssport_info["name"] = m.group(1)
                # Fix naming to match san top name standard
                # use lower case
                ssport_info["name"] = ssport_info["name"].lower()
                # remove spaces
                ssport_info["name"] = ssport_info["name"].replace(" ", "")
                continue

            # Skip while LUN name is not found
            if not ssport_info:
                continue

            m = _spport_port_id_regex.match(info)
            if m:
                if not ssport_info:
                    _print("FAIL: Does not know to which ssport add ID: %s" % m.group(1))
                    continue
                ssport_info["port_id"] = m.group(1)
                continue

            m = _spport_port_uid_regex.match(info)
            if m:
                if not ssport_info:
                    _print("FAIL: Does not know to which ssport add uid: %s" % m.group(1))
                    continue
                ssport_info["uid"] = m.group(1)
                port_info = _uid2wwnn_wwpn(m.group(1))
                if port_info:
                    ssport_info["wwnn"] = port_info["wwnn"]
                    ssport_info["wwpn"] = port_info["wwpn"]

        return self.ssports

    def _lun_exist(self, lun_name):
        """Test if LUN exist
        If LUN does not exist return error
        """
        if not lun_name:
            _print("FAIL: _lun_exist - requires lun_name")
            return False

        luns_dict = self.query_all_lun_info()
        if not luns_dict:
            _print("FAIL: _lun_exist() - Could not query all lun info")
            return False

        if lun_name in list(luns_dict.keys()):
            return True
        return False

    def lun_info(self, lun_name):
        """Query detailed information of specific LUN"""
        if not lun_name:
            _print("FAIL: lun_info - requires lun_name")
            return None

        luns_dict = self.query_all_lun_info()
        if not luns_dict:
            _print("FAIL: lun_info() - Could not query all lun info")
            return None
        if lun_name in list(luns_dict.keys()):
            return luns_dict[lun_name]

        _print("INFO lun_info(): Could not find LUN %s" % lun_name)
        return None

    def lun_create(self, name, size):
        """Create new LUN on array if LUN does not exist
        Purpose
            Perform LUN creation on EMC VNX. This method only create LUN
            on strorage array, you need call lun_map() or lun_create_and_map().

            If $lun_name is undefined, we will use the lowest free LUN ID.

            Before this method return, we will update lun info by call
            query_all_lun_info().
            We need the output of these commands on EMC VNX.
                lun -create type "Thin"
                    -capacity int(size / 512) -sq bc \
                    -poolId ctrl_storage_pool \
                    -name lun_name
        Parameters
            size        # size, accept exapmle: "1KiB" "2GiB" "3MiB"
                        # http://physics.nist.gov/cuu/Units/binary.html
            lun_name    # the LUN name on storage array. like:
                        #  tmp_lun_20160411192741815629
        Return
            lun_name
            or
            None
        """
        size = int(size)
        _print(f"INFO: Creating LUN {name} with size {size}")
        if not name or not size:
            _print("FAIL: usage lun_create(name, size)")
            return None

        # We are always using Thin provisioning
        thinp_str = "Thin"

        if self._lun_exist(name):
            _print("INFO: LUN %s already exist" % name)
            return None

        size_bytes = size

        if not size_bytes:
            _print("FAIL: lun_create() - Could not convert %s to bytes" % size)
            return None

        capacity = int(size / self.BLOCK_SIZE_EMC_VNX_VNX)
        if not capacity or self.ctrl_storage_pool is None:
            _print("FAIL: lun_create() - missing capacity or ctrl storage pool parameter")
            return None

        cmd = "lun -create -type {} -poolId {} -sq bc -name {} -capacity {} -sp a -aa 1".format(
            thinp_str,
            self.ctrl_storage_pool,
            name,
            capacity,
        )

        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: lun_create() - "%s" command' % cmd)
            print(output)
            return None
        # Need to update our volumes
        luns_dict = self.query_all_lun_info(recheck=True)
        if name not in list(luns_dict.keys()):
            self.lun_remove(name)
            _print("FAIL: thought LUN %s was created, but it was not" % name)
            print(list(luns_dict.keys()))
            return None

        _print("INFO: LUN %s created successfully" % name)
        return name

    def lun_remove(self, name):
        """Delete a LUN on array if LUN exists
        Purpose
            Remove LUN on EMC VNX.
            We depend on these commands:
                lun -destroy -name lun_name -o
            If that LUN is mapped, we ill unmap it before we destroy it.
        Parameters
            lun_name
        Returns
            True       #remove pass or no such lun exsit
                or
            False
        """
        if not name:
            _print("FAIL: usage lun_remove(name)")
            return False

        self.query_all_lun_info(recheck=True)
        if not self._lun_exist(name):
            _print("INFO: LUN %s do NOT exist" % name)
            return True

        lun_info_dict = self.lun_info(name)
        if not lun_info_dict:
            _print("FAIL: Could not get information for LUN %s" % name)
            return False

        # Need to unmap mapped groups
        unmap_st_grps = []
        if "map_infos" in list(lun_info_dict.keys()):
            for lun_map_info in lun_info_dict["map_infos"]:
                if lun_map_info["st_group_name"] not in unmap_st_grps:
                    unmap_st_grps.append(lun_map_info["st_group_name"])

        if unmap_st_grps:
            for unmap_grp in unmap_st_grps:
                if not self.lun_unmap(name, st_grp_name=unmap_grp):
                    _print(f"FAIL: lun_remove() - Could not unmap LUN {name} from '{unmap_grp}'")
                    return False

        _print("INFO: Deleting LUN %s" % name)
        cmd = "lun -destroy -name %s -o" % name
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: "%s" command' % cmd)
            print(output)
            return False
        # VNX takes a while to update lun info
        max_wait = 10
        while max_wait:
            libsan.host.linux.sleep(5)
            # Need to update our volumes
            luns_dict = self.query_all_lun_info(recheck=True)
            if name not in list(luns_dict.keys()):
                _print("INFO: LUN %s got removed" % name)
                return True
            _print(f"INFO: LUN {name} still exist, keep waiting {max_wait}")
            max_wait -= 1
        _print("FAIL: Thought LUN %s was deleted, but it is still on array" % name)
        return False

    def lun_map(self, name, lun_id=None):
        """Map a LUN to an igroup"""
        if not name:
            _print("FAIL: usage lun_map(name)")
            return False

        if not self._lun_exist(name):
            _print("INFO: LUN %s do NOT exist" % name)
            return False

        if not self.map_info:
            _print("FAIL: lun_map() - There is no mapping info to use")
            return False

        st_grps_dict = self.query_all_storage_group_info()
        if not st_grps_dict:
            _print("FAIL: lun_map() - Could not find any storage group on VXN array")
            return False

        # To which storage group we should map the LUN
        st_grps2map = []
        # Check to to which storage group the LUN should be mapped
        for map_info in self.map_info:
            if "t_wwpn" in list(map_info.keys()) and "h_wwpn" in list(map_info.keys()):
                for st_grp in list(st_grps_dict.keys()):
                    if "hba_infos" not in list(st_grps_dict[st_grp].keys()):
                        continue
                    # _print("DEBUG: st_grp:")
                    # print st_grps_dict[st_grp]
                    for hba_info in st_grps_dict[st_grp]["hba_infos"]:
                        if (
                            "h_wwpn" in list(hba_info.keys())
                            and hba_info["h_wwpn"] == map_info["h_wwpn"]
                            and st_grps_dict[st_grp]["name"] not in st_grps2map
                        ):
                            # _print("DEBUG: lun_map")
                            # print hba_info
                            st_grps2map.append(st_grps_dict[st_grp]["name"])

            if "t_iqn" in list(map_info.keys()) and "h_iqn" in list(map_info.keys()):
                for st_grp in list(st_grps_dict.keys()):
                    if "hba_infos" not in list(st_grps_dict[st_grp].keys()):
                        continue
                    # _print("DEBUG: st_grp:")
                    # print st_grps_dict[st_grp]
                    for hba_info in st_grps_dict[st_grp]["hba_infos"]:
                        if "h_iqn" in list(hba_info.keys()) and hba_info["h_iqn"] == map_info["h_iqn"]:  # noqa: SIM102
                            if st_grps_dict[st_grp]["name"] not in st_grps2map:
                                # _print("DEBUG: lun_map")
                                # print hba_info
                                st_grps2map.append(st_grps_dict[st_grp]["name"])

            if not st_grps2map:
                _print("FAIL: lun_map() - Do not know how to map using")
                print(map_info)
                _print("Storage groups on VNX array:")
                print(st_grps_dict)
                return False

        lun_info_dict = self.lun_info(name)
        if not lun_info_dict:
            _print("FAIL: Could not query LUN %s info" % name)
            return False

        for my_stgrp in st_grps2map:
            # print st_grps_dict[my_stgrp]
            stgrp_mapped = False
            if "map_infos" in list(lun_info_dict.keys()):
                for lun_map_info in lun_info_dict["map_infos"]:
                    # Just skip print info multiple times
                    if stgrp_mapped:
                        continue
                    if lun_map_info["st_group_name"] == my_stgrp:
                        # LUN already mapped to this storage group
                        _print(f"INFO: LUN {name} is already mapped to '{my_stgrp}'")
                        stgrp_mapped = True
                        continue
            if stgrp_mapped:
                continue
            # Search for lowest lun_id it can assign to host
            if lun_id is None:
                _lun_ids = []
            if "map_infos" in list(st_grps_dict[my_stgrp].keys()):
                for map_info in st_grps_dict[my_stgrp]["map_infos"]:
                    if "h_lun_id" in list(map_info.keys()):
                        _lun_ids.append(map_info["h_lun_id"])

            lun_id = libsan.misc.array.lowest_free_number(_lun_ids)
            # _print("DEBUG: lun_map() - map st_grp %s h_lun_id %s t_lun_id %s" % (my_stgrp, lun_id,
            # lun_info_dict["lun_id"]))
            _print(f"INFO: Mapping LUN {name} to '{my_stgrp}'")
            cmd = "storagegroup -addhlu -gname '{}' -hlu {} -alu {}".format(my_stgrp, lun_id, lun_info_dict["lun_id"])
            ret, output = self._run(cmd, return_output=True, verbose=False)
            if ret != 0:
                _print('FAIL: "%s" command' % cmd)
                print(output)
                return False

        # make sure all h_wwpns that we need to map are mapped
        self.query_all_lun_info(recheck=True)
        lun_info_dict = self.lun_info(name)
        if not lun_info_dict:
            _print("FAIL: lun_map() - Could not query LUN %s" % name)
            return False
        if "map_infos" not in list(lun_info_dict.keys()):
            _print("FAIL: lun_map() - Could not find map info for LUN %s" % name)
            return False

        lun_map_infos = lun_info_dict["map_infos"]
        # _print("DEBUG: lun_map() - found lun maps:")
        # print lun_map_infos
        # print lun_info_dict

        found_map = False
        h_wwpn = None
        for map_info in self.map_info:
            for lun_map_info in lun_map_infos:
                if "h_wwpn" in list(map_info.keys()):
                    h_wwpn = map_info["h_wwpn"]
                    if h_wwpn in lun_map_info["h_wwpn"]:
                        found_map = True
                if "h_iqn" in list(map_info.keys()):
                    h_iqn = map_info["h_iqn"]
                    if h_iqn in lun_map_info["h_iqn"]:
                        found_map = True
            if not found_map:
                _print(f"FAIL: Tought it mapped {h_wwpn} to lun {name}, but it did not")
                return False

        return True

    def lun_unmap(self, name, st_grp_name=None):
        """Unmap a LUN to an IQN"""
        if not name:
            _print("FAIL: usage lun_unmap(name)")
            return False

        if not self._lun_exist(name):
            _print("INFO: lun_unmap() - LUN %s do NOT exist" % name)
            return False

        if not self.map_info and not st_grp_name:
            _print("FAIL: lun_unmap() - There is no mapping info to use")
            return False

        st_grps_dict = self.query_all_storage_group_info()

        # To which igroups we should map the LUN
        st_grps2map = []
        if st_grp_name:
            st_grps2map.append(st_grp_name)
        else:
            # Check to to which storage group the LUN should be mapped
            for map_info in self.map_info:
                if "t_wwpn" in list(map_info.keys()) and "h_wwpn" in list(map_info.keys()):
                    for st_grp in list(st_grps_dict.keys()):
                        if "hba_infos" not in list(st_grps_dict[st_grp].keys()):
                            continue
                        # _print("DEBUG: st_grp:")
                        # print st_grps_dict[st_grp]
                        for hba_info in st_grps_dict[st_grp]["hba_infos"]:
                            if (
                                "h_wwpn" in list(hba_info.keys())
                                and hba_info["h_wwpn"] == map_info["h_wwpn"]
                                and st_grps_dict[st_grp]["name"] not in st_grps2map
                            ):
                                # _print("DEBUG: lun_map")
                                # print hba_info
                                st_grps2map.append(st_grps_dict[st_grp]["name"])

                if "t_iqn" in list(map_info.keys()) and "h_iqn" in list(map_info.keys()):
                    for st_grp in list(st_grps_dict.keys()):
                        if "hba_infos" not in list(st_grps_dict[st_grp].keys()):
                            continue
                        # _print("DEBUG: st_grp:")
                        # print st_grps_dict[st_grp]
                        for hba_info in st_grps_dict[st_grp]["hba_infos"]:
                            if (
                                "h_iqn" in list(hba_info.keys())
                                and hba_info["h_iqn"] == map_info["h_iqn"]
                                and st_grps_dict[st_grp]["name"] not in st_grps2map
                            ):
                                # _print("DEBUG: lun_map")
                                # print hba_info
                                st_grps2map.append(st_grps_dict[st_grp]["name"])

                if not st_grps2map:
                    _print("FAIL: lun_map() - Do not know how to map using")
                    print(map_info)
                    return False

        lun_info_dict = self.lun_info(name)
        if not lun_info_dict:
            _print("FAIL: lun_unmap() - Could not query LUN %s" % name)
            return False
        if "map_infos" not in list(lun_info_dict.keys()):
            _print("INFO: lun_unmap() - LUN %s is not mapped to any storage group" % name)
            return True

        for my_stgrp in st_grps2map:
            # Check if lun is mapped to this group
            unmapped_st_grp = False
            for lun_map_info in lun_info_dict["map_infos"]:
                if my_stgrp != lun_map_info["st_group_name"]:
                    continue
                # This storage group have already been removed
                # for example, storage group is mapped to different ports
                if unmapped_st_grp:
                    continue
                # _print("DEBUG: lun_unmap() - unmap st_grp '%s' h_lun_id %s" % (my_stgrp, lun_map_info["h_lun_id"]))
                # print st_grps_dict[my_stgrp]
                _print(f"INFO: Unmapping LUN {name} from '{my_stgrp}'")
                cmd = "storagegroup -removehlu -gname '{}' -hlu {} -o".format(my_stgrp, lun_map_info["h_lun_id"])
                ret, output = self._run(cmd, return_output=True, verbose=False)
                if ret != 0:
                    _print('FAIL: "%s" command' % cmd)
                    print(output)
                    return False
                unmapped_st_grp = True

        # make sure all h_wwpns that we need to map are mapped
        luns_dict = self.query_all_lun_info(recheck=True)
        if "map_infos" not in list(luns_dict[name].keys()):
            # LUN is not mapped to any initiator
            return True
        lun_map_infos = luns_dict[name]["map_infos"]
        found_map = False
        for map_info in self.map_info:
            for lun_map_info in lun_map_infos:
                if "h_wwpn" in list(map_info.keys()):
                    h_wwpn = map_info["h_wwpn"]
                    if h_wwpn in lun_map_info["h_wwpn"]:
                        found_map = True
                if "h_iqn" in list(map_info.keys()):
                    h_iqn = map_info["h_iqn"]
                    if h_iqn in lun_map_info["h_iqn"]:
                        found_map = True
            if found_map:
                _print("FAIL: Tought it unmapped lun %s, but it did not" % name)
                print(map_info)
                return False

        return True

    # ####### Storage Groups ##############

    def query_all_storage_group_info(self, recheck=False):
        """Query all Storage Group info on array and store its information on a dict"""

        # If we do not need to recheck and port info exist we return it
        # if self.storage_groups_dict and not recheck:
        #    return self.storage_groups_dict

        self.storage_groups_dict = {}

        # If not recheck, it will return exiting information
        getall_info = self._get_all(recheck=recheck)
        if not getall_info:
            return None

        ssport_list = self.query_all_spport_info()

        _storage_grp_section_regex = re.compile(r"All Storage Groups Information")
        _raid_section_regex = re.compile(r"All RAID Groups Information")
        # VNX accepts group name with spaces...
        _storage_grp_name_regex = re.compile(r"Storage Group Name:\s+(.*)$")
        _storage_grp_hba_sp_regex = re.compile(r"HBA/SP Pairs:")
        _storage_grp_hli_alu_regex = re.compile(r"HLU/ALU Pairs:")

        flag_all_storage_grp_info_begin = False
        flag_initiators_begin = False
        flag_map_begin = False
        st_grp_info = None

        for info in getall_info:
            # _print("DEBUG: query_all_storage_group_info() %s" % info)
            if _storage_grp_section_regex.match(info):
                flag_all_storage_grp_info_begin = True
                continue
            if _raid_section_regex.match(info):
                flag_all_storage_grp_info_begin = False
                continue

            if not flag_all_storage_grp_info_begin:
                continue

            # _print("DEBUG: Storage Grp info '%s'" % info)

            m = _storage_grp_name_regex.match(info)
            if m:
                self.storage_groups_dict[m.group(1)] = {}
                st_grp_info = self.storage_groups_dict[m.group(1)]
                st_grp_info["name"] = m.group(1)
                flag_initiators_begin = False
                # _print("DEBUG: Storage Grp name: %s" % m.group(1))
                continue

            if _storage_grp_hba_sp_regex.match(info):
                flag_initiators_begin = True
                continue

            if flag_initiators_begin:
                # _print("DEBUG: query_all_storage_group_info() - Initiator info '%s'" % info)
                m = re.match(r"\s+(\S+)\s+(SP \S+)\s+(\S+)$", info)
                # _print("DEBUG: query_all_storage_group_info() - process initiators %s" % info)
                if m:
                    if not st_grp_info:
                        _print(
                            "FAIL: query_all_storage_group_info() - Does not know to which storage group add HBA "
                            "info: %s" % info
                        )
                        continue
                    hba_info = {}
                    wwn_info = _uid2wwnn_wwpn(m.group(1))
                    if wwn_info:
                        hba_info["h_wwpn"] = wwn_info["wwpn"]
                    elif libsan.host.iscsi.is_iqn(m.group(1)):
                        hba_info["h_iqn"] = m.group(1)
                    hba_info["sp_name"] = m.group(2)
                    hba_info["sp_port"] = m.group(3)
                    # Check wwpn of this port
                    if ssport_list:
                        for ssport in ssport_list:
                            if (
                                "port_id" in list(ssport.keys())
                                and ssport["name"] == hba_info["sp_name"]
                                and ssport["port_id"] == hba_info["sp_port"]
                            ):
                                wwn_info = _uid2wwnn_wwpn(ssport["uid"])
                                if wwn_info:
                                    hba_info["t_wwpn"] = ssport["wwpn"]
                                elif libsan.host.iscsi.is_iqn(ssport["uid"]):
                                    hba_info["t_iqn"] = ssport["uid"]
                    if "hba_infos" not in list(st_grp_info.keys()):
                        st_grp_info["hba_infos"] = []
                    # Check if entry already exist
                    found_info = False
                    for h_info in st_grp_info["hba_infos"]:
                        if (
                            h_info["sp_name"] == hba_info["sp_name"]
                            and h_info["sp_port"] == hba_info["sp_port"]
                            and (
                                "h_wwpn" in list(hba_info.keys())
                                and h_info["h_wwpn"] == hba_info["h_wwpn"]
                                or "h_iqn" in list(hba_info.keys())
                                and h_info["h_iqn"] == hba_info["h_iqn"]
                            )
                        ):
                            found_info = True
                    if not found_info:
                        st_grp_info["hba_infos"].append(hba_info)

            if _storage_grp_hli_alu_regex.match(info):
                flag_map_begin = True
                flag_initiators_begin = False
                continue
            # If line has : this is part of some other information
            if re.search(":", info):
                flag_map_begin = False
                continue

            if flag_map_begin:
                # _print("DEBUG: query_all_storage_group_info() - Map info '%s'" % info)
                m = re.match(r"\s+(\d+)\s+(\d+)", info)
                if m:
                    if not st_grp_info:
                        _print(
                            "FAIL: query_all_storage_group_info() - Does not know to which storage group add "
                            "map: {} - {}".format(m.group(1), m.group(2))
                        )
                        continue
                    map_info = {"h_lun_id": m.group(1), "t_lun_id": m.group(2)}
                    if "map_infos" not in list(st_grp_info.keys()):
                        st_grp_info["map_infos"] = []
                    if map_info not in st_grp_info["map_infos"]:
                        st_grp_info["map_infos"].append(map_info)

        return self.storage_groups_dict

    # ####### Controllers ##############
    def sa_ctrler_t_wwpns(self, ctrler_ip):
        """Usage
            self.sa_ctrler_t_wwpns(ctrler_ip)
        Parameter
            ctrler_ip    # like 10.10.1.2
        Purpose
            Login to controller and list t_wwpns in the controller
        Returns
            A list of t_wwpns
        """

        t_wwpns = []

        ssport_list = self.query_all_spport_info()

        if not ssport_list:
            return None
        ctrler_name_regex = re.compile(r"ctrl_ip_(\S+)")
        ctrler_name = None
        if self.sa_conf_dict:
            # Search if there is a controller name defined for this controller
            for key in list(self.sa_conf_dict.keys()):
                if self.sa_conf_dict[key] == ctrler_ip:
                    m = ctrler_name_regex.match(key)
                    if m:
                        ctrler_name = m.group(1)

        if not ctrler_name:
            _print("FAIL: Could not find controller name configured for %s" % ctrler_ip)
            print(self.sa_conf_dict)
            return None

        for ssport in ssport_list:
            if ssport["name"] == ctrler_name and "wwpn" in list(ssport.keys()) and ssport["wwpn"] not in t_wwpns:
                t_wwpns.append(ssport["wwpn"])

        return t_wwpns

    def sa_ctrler_reboot(self, ctrler_ip):
        """Usage
            self.sa_ctrler_reboot(ctrler_ip)
        Purpose
            Reboot certain controller. This function does not wait controller become
            online.
        Parameter
            ctrler_ip    # like 10.10.1.2
        Returns
            True
            or
            False
        """

        cmd = "rebootSP -o"
        ret, output = self._run(cmd, ctrler=ctrler_ip, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: sa_ctrler_reboot() "%s" command' % cmd)
            print(output)
            return False

        # Let's check if controller is down
        timeout = 120
        interval = 10
        loop_count = timeout / interval
        while loop_count > 1:
            loop_count -= 1
            status = self.sa_ctrler_check(ctrler_ip)
            if not status or status == "offline":
                return True
            libsan.host.linux.sleep(interval)

        _print("FAIL: sa_ctrler_reboot() - Controller has not rebooted")
        print(output)
        return False

    def sa_ctrler_check(self, ctrler_ip):
        """Usage
        self.sa_ctrler_check(ctrler_name)
        Purpose
            Check whether controller is online or not. We depend on this command:
                getagent
        Parameter
            ctrler_ip    # like 10.10.1.2
        Returns
            ctrler_status  # 'online' or 'offline'
            or
            None
        """

        status = "offline"
        cmd = "getagent"
        ret, output = self._run(cmd, ctrler=ctrler_ip, return_output=True, verbose=False)
        if ret != 0:
            return status

        if not output:
            return "offline"
        lines = output.split("\n")
        online_regex = re.compile(r"Agent Rev:")
        for line in lines:
            if online_regex.search(line):
                status = "online"
        return status



#!/usr/bin/python
import pytest

# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.
from libsan import sanmgmt


def test_load_san_config():
    san_obj = sanmgmt.SanMgmt()
    if not san_obj.load_conf():
        pytest.fail("Unable to load config")

    for hw_entry in san_obj.san_conf_dict:
        hw_info = san_obj.san_conf_dict[hw_entry]
        if "class_name" in hw_info:
            # Try to load class
            if not sanmgmt.get_class(hw_info["module_name"], hw_info["class_name"]):
                pytest.fail("Unable to load class")
            print("loaded class from {}.{}".format(hw_info["module_name"], hw_info["class_name"]))

    assert 1



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""lvm.py: Module to manipulate LVM devices."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import fileinput
import re
from functools import wraps

from libsan import _print
from libsan.host.cli_tools import Wrapper
from libsan.host.cmdline import run
from libsan.host.vdo import VDO


###########################################
# PV section
###########################################
def pv_query(verbose=False):
    """Query Physical Volumes and return a dictonary with PV information for each PV.
    The arguments are:
    \tNone
    Returns:
    \tdict: Return a dictionary with PV info for each PV
    """
    cmd = 'pvs --noheadings --separator ","'
    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("INFO: there is no VGs")
        return None
    pvs = output.split("\n")

    # format of PV info: PV,VG,Fmt,Attr,PSize,PFree
    pv_info_regex = r"\s+(\S+),(\S+)?,(\S+),(.*),(.*),(.*)$"

    pv_dict = {}
    for pv in pvs:
        m = re.match(pv_info_regex, pv)
        if not m:
            # _print("WARN: (%s) does not match vgdisplay output format" % vg)
            continue
        pv_info_dict = {
            "vg": m.group(2),
            "fmt": m.group(3),  # not sure what it is
            "attr": m.group(4),
            "psize": m.group(5),
            "pfree": m.group(6),
        }
        pv_dict[m.group(1)] = pv_info_dict

    return pv_dict


def pv_create(pv_name: str, options="", verbose=True):
    """Create a Volume Group.
    The arguments are:
    \tPV name
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    if not pv_name:
        _print("FAIL: pv_create requires pv_name")
        return False
    cmd = f"pvcreate {options} {pv_name}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        # _print ("FAIL: Could not create %s" % pv_name)
        return False
    return True


def pv_remove(pv_name: str, force=None, verbose=True):
    """Delete a Volume Group.
    The arguments are:
    \tVG name
    \tforce (boolean)
    \tverbose (boolean)
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t\tFalse in case of failure
    """
    if not pv_name:
        _print("FAIL: pv_remove requires pv_name")
        return False

    pv_dict = pv_query()

    pv_names = pv_name.split()
    for pv_name in pv_names:
        if pv_name not in list(pv_dict.keys()):
            _print("INFO: pv_remove - %s does not exist. Skipping..." % pv_name)
            return True

        options = ""
        if force:
            options += "--force --force"
        cmd = f"pvremove {options} {pv_name}"
        retcode = run(cmd, verbose=verbose)
        if retcode != 0:
            _print("FAIL: Could not delete %s" % pv_name)
            return False
    return True


###########################################
# VG section
###########################################


def vg_show():
    """Show information for Volume Groups
    The arguments are:
    \tNone
    Returns:
    \tTrue
    or
    \tFalse
    """
    cmd = "vgs -a"
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: Could not show VGs")
        return False
    return True


def vg_query(verbose=False):
    """Query Volume Groups and return a dictonary with VG information for each VG.
    The arguments are:
    \tNone
    Returns:
    \tdict: Return a dictionary with VG info for each VG
    """
    cmd = 'vgs --noheadings --separator ","'
    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("INFO: there is no VGs")
        return None
    vgs = output.split("\n")

    # format of VG info: name #PV #LV #SN Attr VSize VFree
    vg_info_regex = r"\s+(\S+),(\S+),(\S+),(.*),(.*),(.*),(.*)$"

    vg_dict = {}
    for vg in vgs:
        m = re.match(vg_info_regex, vg)
        if not m:
            # _print("WARN: (%s) does not match vgdisplay output format" % vg)
            continue
        vg_info_dict = {
            "num_pvs": m.group(2),
            "num_lvs": m.group(3),
            "num_sn": m.group(4),  # not sure what it is
            "attr": m.group(5),
            "vsize": m.group(6),
            "vfree": m.group(7),
        }
        vg_dict[m.group(1)] = vg_info_dict

    return vg_dict


def vg_create(vg_name: str, pv_name: str, force=False, verbose=True):
    """Create a Volume Group.
    The arguments are:
    \tPV name
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """
    if not vg_name or not pv_name:
        _print("FAIL: vg_create requires vg_name and pv_name")
        return False

    options = ""
    if force:
        options += "--force"
    cmd = f"vgcreate {options} {vg_name} {pv_name}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        # _print ("FAIL: Could not create %s" % vg_name)
        return False
    return True


def vg_remove(vg_name: str, force=False, verbose=True):
    """Delete a Volume Group.
    The arguments are:
    \tVG name
    \tforce (boolean)
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """
    if not vg_name:
        _print("FAIL: vg_remove requires vg_name")
        return False

    vg_dict = vg_query()
    if vg_name not in list(vg_dict.keys()):
        _print("INFO: vg_remove - %s does not exist. Skipping..." % vg_name)
        return True

    options = ""
    if force:
        options += "--force"
    cmd = f"vgremove {options} {vg_name}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        # _print ("FAIL: Could not delete %s" % vg_name)
        return False
    return True


###########################################
# LV section
###########################################


def lv_show():
    """Show information for Logical Volumes
    The arguments are:
    \tNone
    Returns:
    \tTrue
    or
    \tFalse
    """
    cmd = "lvs -a"
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: Could not show LVs")
        return False
    return True


def lv_query(options="", verbose=False):
    """Query Logical Volumes and return a dictonary with LV information for each LV.
    The arguments are:
    \toptions:  If not want to use default lvs output. Use -o for no default fields
    Returns:
    \tdict: Return a list with LV info for each LV
    """
    # Use \",\" as separator, as some output might contain ','
    # For example, lvs -o modules on thin device returns "thin,thin-pool"
    cmd = 'lvs -a --noheadings --separator \\",\\"'

    # format of LV info: Name VG Attr LSize Pool Origin Data%  Meta%  Move Log Cpy%Sync Convert
    lv_info_regex = (
        r"\s+(\S+)\",\"(\S+)\",\"(\S+)\""
        r",\"(\S+)\",\"(.*)\",\"(.*)\",\"(.*)\",\"(.*)\",\"(.*)\",\"(.*)\",\"(.*)\",\"(.*)$"
    )

    # default parameters returned by lvs -a
    param_names = [
        "name",
        "vg_name",
        "attr",
        "size",
        "pool",
        "origin",
        "data_per",
        "meta_per",
        "move",
        "log",
        "copy_per",
        "convert",
    ]

    if options:
        param_names = ["name", "vg_name"]
        # need to change default regex
        lv_info_regex = r"\s+(\S+)\",\"(\S+)"
        parameters = options.split(",")
        for param in parameters:
            lv_info_regex += '","(.*)'
            param_names.append(param)
        lv_info_regex += "$"
        cmd += " -o lv_name,vg_name,%s" % options

    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("INFO: there is no LVs")
        return None
    lvs = output.split("\n")

    lv_list = []
    for lv in lvs:
        m = re.match(lv_info_regex, lv)
        if not m:
            _print("FAIL: (%s) does not match lvs output format" % lv)
            continue
        lv_info_dict = {}
        for index in range(len(param_names)):
            lv_info_dict[param_names[index]] = m.group(index + 1)
        lv_list.append(lv_info_dict)

    return lv_list


def lv_create(vg_name: str, lv_name: str, options: list, verbose=True):
    """Create a Logical Volume.
    The arguments are:
    \tVG name
    \tLV name
    \toptions
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """
    if not vg_name or not lv_name:
        _print("FAIL: lv_create requires vg_name and lv_name")
        return False

    cmd = "lvcreate {} {} -n {}".format(" ".join(str(i) for i in options), vg_name, lv_name)
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        # _print ("FAIL: Could not create %s" % lv_name)
        return False
    return True


def lv_info(lv_name: str, vg_name: str, options="", verbose=False):
    """Show information of specific LV"""
    if not lv_name or not vg_name:
        _print("FAIL: lv_info() - requires lv_name and vg_name as parameters")
        return None

    lvs = lv_query(options=options, verbose=verbose)

    if not lvs:
        return None

    for lv in lvs:
        if lv["name"] == lv_name and lv["vg_name"] == vg_name:
            return lv
    return None


def lv_activate(lv_name: str, vg_name: str, verbose=True):
    """Activate a Logical Volume
    The arguments are:
    \tLV name
    \tVG name
    Returns:
    \tBoolean:
    \t\tTrue in case of success
    \t\tFalse if something went wrong
    """
    if not lv_name or not vg_name:
        _print("FAIL: lv_activate requires lv_name and vg_name")
        return False

    cmd = f"lvchange -ay {vg_name}/{lv_name}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not activate LV %s" % lv_name)
        return False

    # Maybe we should query the LVs and make sure it is really activated
    return True


def lv_deactivate(lv_name: str, vg_name: str, verbose=True):
    """Deactivate a Logical Volume
    The arguments are:
    \tLV name
    \tVG name
    Returns:
    \tBoolean:
    \t\tTrue in case of success
    \t\tFalse if something went wrong
    """
    if not lv_name or not vg_name:
        _print("FAIL: lv_deactivate requires lv_name and vg_name")
        return False

    cmd = f"lvchange -an {vg_name}/{lv_name}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not deactivate LV %s" % lv_name)
        return False

    # Maybe we should query the LVs and make sure it is really deactivated
    return True


def lv_remove(lv_name: str, vg_name: str, verbose=True):
    """Remove an LV from a VG
    The arguments are:
    \tLV name
    \tVG name
    Returns:
    \tBoolean:
    \t\tTrue in case of success
    \t\tFalse if something went wrong
    """
    if not lv_name or not vg_name:
        _print("FAIL: lv_remove requires lv_name and vg_name")
        return False

    lv_names = lv_name.split()

    for lv_name in lv_names:
        if not lv_info(lv_name, vg_name):
            _print("INFO: lv_remove - LV %s does not exist. Skipping" % lv_name)
            continue

        cmd = f"lvremove --force {vg_name}/{lv_name}"
        retcode = run(cmd, verbose=verbose)
        if retcode != 0:
            _print("FAIL: Could not remove LV %s" % lv_name)
            return False

        if lv_info(lv_name, vg_name):
            _print("INFO: lv_remove - LV %s still exists." % lv_name)
            return False

    return True


def lv_convert(vg_name: str, lv_name: str, options: list, verbose=True):
    """Change Logical Volume layout.
    The arguments are:
    \tVG name
    \tLV name
    \toptions
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """
    if not options:
        _print("FAIL: lv_convert requires at least some options specified.")
        return False

    if not lv_name or not vg_name:
        _print("FAIL: lv_convert requires vg_name and lv_name")
        return False

    cmd = "lvconvert {} {}/{}".format(" ".join(options), vg_name, lv_name)
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not convert %s" % lv_name)
        return False

    return True


def lv_extend(vg_name: str, lv_name: str, options: list, verbose=True):
    """Increase size of logical volume.
    The arguments are:
    \tVG name
    \tLV name
    \toptions
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """
    if not options:
        _print("FAIL: lv_extend requires at least some options specified.")
        return False

    if not lv_name or not vg_name:
        _print("FAIL: lv_extend requires vg_name and lv_name")
        return False

    cmd = "lvextend {} {}/{}".format(" ".join(options), vg_name, lv_name)
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not extend %s" % lv_name)
        return False

    return True


def lv_reduce(vg_name: str, lv_name: str, options: list, verbose=True):
    """Decrease size of logical volume.
    The arguments are:
    \tVG name
    \tLV name
    \toptions
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """
    if not options:
        _print("FAIL: lv_reduce requires at least some options specified.")
        return False

    if not lv_name or not vg_name:
        _print("FAIL: lv_reduce requires vg_name and lv_name")
        return False

    cmd = "lvreduce {} {}/{}".format(" ".join(options), vg_name, lv_name)
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not reduce %s" % lv_name)
        return False

    return True


###########################################
# Config file
###########################################


def get_config_file_path():
    return "/etc/lvm/lvm.conf"


def update_config(key: str, value: str):
    config_file = get_config_file_path()
    search_regex = re.compile(r"(\s*)%s(\s*)=(\s*)\S*" % key)
    search_regex_with_comment = re.compile(r"(\s*#\s*)%s(\s*)=(\s*)\S*" % key)
    for line in fileinput.input(config_file, inplace=True):
        m = search_regex.match(line)
        m_with_comment = search_regex_with_comment.match(line)
        if m:
            line = f"{m.group(1)}{key} = {value}"
        if m_with_comment:
            line = "{}{} = {}".format(m_with_comment.group(1).replace("#", ""), key, value)
        # print saves the line to the file
        # need to remove new line character as print will add it
        line = line.rstrip("\n")
        print(line)


def get_lvm_config_options(all_params=False):
    """Get all the configuration types from lvm.conf file"""
    _, out = run("lvmconfig --type full", return_output=True, verbose=False)

    options = {}
    category = ""
    for line in out.split("\n"):
        line = line.strip()
        if line == "" or "}" in line:
            # skip empty or end of list line
            continue
        if "{" in line:
            # category part
            category = line[:-2]  # removing "allocation {"[:-2] == "allocation"
            options[category] = []
        elif "=" in line:
            # content of a category
            options[category].append(line.split("=")[0])

    if not all_params:
        return options

    options_all = []
    for value in options.values():
        options_all.extend(value)
    return options_all


def check_lvm_config(option):
    _, out = run(f"lvs -o {option}", return_output=True)
    return out


def get_all_lvm_config_options():
    all_opts = check_lvm_config("asdf")  # needs just something that it doesn't know
    # we need to filter on lines that look like this
    # lvm_blahblah - explanation
    # and remove separators

    # at the end is '?' which should not be there therefore [:-1]
    return [i.split()[0] for i in all_opts if re.match(r".+-.+", i) and "--" not in i][:-1]


def run_cmd(func):
    # TODO: Duplicate of run_command
    """Decorator for running commands
    kwargs need to be edited every time, so decorator is probably the best solution
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        return run(func(), *args, **kwargs, return_output=True)  # cmd == func()

    return wrapped


@run_cmd
def get_all_lvmvdoconfig_options():
    return "lvmconfig --type list"


@run_cmd
def lvdisplay():
    return "lvdisplay"


def print_profile_file(profile_name, path=None):
    if not path:
        path = "/etc/lvm/profile"  # default profile location
    run(f"cat {path}/{profile_name}.profile")


def get_lvdisplay_data():
    _, out = lvdisplay()
    try:
        lines = [line for line in out if "LV Name" in line and "pool" not in line][0].split()
    except Exception as exc:
        lines = [""]
        print("Exception", exc)
    print(lines)
    return lines[-1]


def run_command(func):
    """Decorator for running commands
    kwargs need to be edited every time, so decorator is probably the best solution
    """

    @wraps(func)
    def wrapped(inst, **kwargs):
        inst.tmp_kwargs = kwargs
        # The first thing is to replace values that are
        # conf=fmf_conf_value -> conf=conf_value
        kwargs = inst.replace_multiple_option_values(**kwargs)
        # check configuration arguments from old conf to new
        # slab_size=minimum -> slab_size=compute(value)
        kwargs = inst.check_config_arguments(**kwargs)

        # create command
        cmd, kwargs = func(inst, **kwargs)

        # remove everything not necessary
        kwargs = inst.remove_nones(inst.remove_vdo_arguments(**kwargs))

        # Check
        if inst.check(**kwargs) is not True:  # check() with "the manual"
            print("Check failed.")
            return False

        # Run
        return inst.run(cmd, **kwargs)

    return wrapped


class LVM(Wrapper):
    """Class for creating LVMVDO commands"""

    def __init__(self, disable_check=True, vdo_arguments_flag=True):
        self.disable_check = disable_check
        self.tmp_kwargs = {}

        self.commands = {
            "lvcreate": "lvcreate",
            "lvremove": "lvremove",
            "lvconvert": "lvconvert",
            "lvchange": "lvchange",
            "lvextend": "lvextend",
            "lvreduce": "lvreduce",
            "lvresize": "lvresize",
        }

        self.commands["all"] = list(self.commands.keys())

        self.arguments = {  # "": [[""], "--"],
            "help": [self.commands["all"], " --help"],
            "version": [self.commands["all"], " --version"],
            "verbose": [self.commands["all"], " --verbose"],
            "type": [["lvcreate", "lvconvert"], "--type&"],
            "vdo": [["lvcreate"], "--vdo"],
            "vdopool": [["lvcreate"], "--vdopool&"],
            "vdo_name": [["lvcreate", "lvconvert"], "--name&"],
            "size": [["lvcreate", "lvextend"], "--size&"],
            "logical_size": [["lvcreate", "lvconvert"], "--virtualsize&"],
            "config": [["lvcreate", "lvextend"], "--config&"],
            "extents": [["lvcreate"], "--extents&"],
            "stripes": [["lvcreate", "lvextend"], "--stripes&"],
            "stripesize": [["lvcreate", "lvextend"], "--stripesize&"],
            "compression": [["lvcreate", "lvconvert", "lvchange"], "--compression&"],
            "deduplication": [["lvcreate", "lvconvert", "lvchange"], "--deduplication&"],
            "activate": [["lvchange"], "--activate&"],
            "metadataprofile": [["lvcreate", "lvconvert", "lvchange"], "--metadataprofile&"],
            "force": [["lvremove"], "--force"],
            "yes": [["lvremove"], "-y"],
            "refresh": [["lvchange"], ""],
            "vg_name": [["lvcreate", "lvextend"], ""],
            "lv_name": [["lvchange"], ""],
        }

        self.no_check_size_unit = [
            "logical_size",
            "cachesize",
            "chunksize",
            "poolmetadatasize",
            "regionsize",
            "size",
            "stripesize",
            "virtualsize",
        ]

        self.argument_options = {
            # argument: [(.fmf option name values), (command option values)]
            # order in tuples matters
            # enable == 'y', disable == 'n'
            "activate": [("enabled", "disabled", "auto"), ("y", "n", "ay")],
            "deduplication": [("enabled", "disabled"), ("y", "n")],
            "compression": [("enabled", "disabled"), ("y", "n")]
            # '': [(,),(,)],
        }

        self.multiple_option_arguments = self.argument_options.keys()

        self.lvm_arguments = self.arguments.copy()

        self.option_translator_dictionary = {
            "1": "vdo_use_compression",
            "2": "vdo_use_deduplication",
            "3": "vdo_use_metadata_hints",
            "4": "vdo_minimum_io_size",
            "block_map_cache_size": "vdo_block_map_cache_size_mb",
            "block_map_period": "vdo_block_map_period",
            "5": "vdo_check_point_frequency",
            "sparse_index": "vdo_use_sparse_index",
            "index_mem": "vdo_index_memory_size_mb",
            "slab_size": "vdo_slab_size_mb",
            "ack_threads": "vdo_ack_threads",
            "bio_threads": "vdo_bio_threads",
            "bio_rotation_interval": "vdo_bio_rotation",
            "cpu_threads": "vdo_cpu_threads",
            "hash_zone_threads": "vdo_hash_zone_threads",
            "logical_threads": "vdo_logical_threads",
            "physical_threads": "vdo_physical_threads",
            "write_policy": "vdo_write_policy",
            "max_discard_size": "vdo_max_discard",
            "6": "vdo_pool_header_size",
        }
        self.config_arguments = self.option_translator_dictionary.keys()

        self.vdo_arguments = {}
        if vdo_arguments_flag:
            # obtain arguments from VDO for backwards compatibility
            self.vdo_arguments = VDO().arguments.copy()

            # cycle through keys common for both VDO and LVMVDO
            for vdo_arg in [arg for arg in self.vdo_arguments if arg in self.lvm_arguments]:
                self.vdo_arguments.pop(vdo_arg)  # pop so they don't get overwritten
            self.arguments.update(self.vdo_arguments)

        self.config_options = get_lvm_config_options()
        self.all_config_options = get_lvm_config_options(all_params=True)

        # backward compatibility between vdo and lvmvdo packages,
        # so that commands correspond to each other
        self.create = self.lvcreate

        Wrapper.__init__(self, self.commands, self.arguments, self.disable_check)

    @staticmethod
    def remove_nones(kwargs):
        return {k: v for k, v in kwargs.items() if v is not None}

    @staticmethod
    def _add_value(value, command, argument):
        """copied from Wrapper"""
        if argument[-1:] in ["=", "&"]:
            if argument[-1:] == "&":
                argument = argument[:-1] + " "
            if isinstance(value, list):
                # allows to use repeatable arguments as a list of values
                for val in value:
                    command += argument + "'" + str(val) + "'"
            else:
                command += argument + "'" + str(value) + "'"
        elif argument[-1:] in "*":
            command += str(value)
        else:
            command += argument
        return command + " "  # added space at the end to fix formatting

    def _add_argument(self, arg, value, command):
        """copied from Wrapper"""
        # Checks if given argument is allowed for given command and adds it to cmd string
        self._check_allowed_argument(arg, command)
        command = self._add_value(value, command, self._get_arg(arg))
        return command

    def _add_arguments(self, cmd, **kwargs):
        """copied from Wrapper"""
        command = cmd
        for kwarg in kwargs:
            # skip adding this argument if the value is False
            if kwargs[kwarg] is False:
                continue
            # skip:
            # vg_name, it should be last
            # vdo_name should be just "name" but not used is OK :D
            if kwarg in ("vg_name"):  # ,"vdo_name"):  # maybe more names will be here
                continue
            command = self._add_argument(kwarg, kwargs[kwarg], command)

        # vg_name must be the last in command
        if "vg_name" in kwargs:
            command += kwargs["vg_name"]
        return command

    def _get_possible_arguments(self, command=None):
        return super()._get_possible_arguments(command.split()[0])

    def run(self, cmd, verbosity=True, return_output=False, **kwargs):
        """Constructs the command to run and runs it"""
        cmd = self._add_arguments(cmd, **kwargs)

        print(cmd)
        ret = run(cmd, verbose=verbosity, return_output=return_output)

        if isinstance(ret, tuple) and ret[0] != 0:
            _print("WARN: Running command: '%s' failed. Return with output." % cmd)
        elif isinstance(ret, int) and ret != 0:
            _print("WARN: Running command: '%s' failed." % cmd)
        return ret

    def number_of_conf_args(self, **kwargs):
        """Counts the number of configuration arguments in kwargs and returns it"""
        return sum(conf_arg in kwargs for conf_arg in self.config_arguments)

    def _set_kwargs_value(self, arg, **kwargs):
        """If the value in test declaration is 'enabled'
        it switches it to the one in the man page, e.g. 'y'
        self.argument_options['activate']
        """
        fmf_options = self.argument_options[arg][0]
        opt_values = self.argument_options[arg][1]

        print(fmf_options)
        for idx, value in enumerate(fmf_options):
            kwargs[arg] = opt_values[idx] if kwargs[arg] == value else kwargs[arg]

        return kwargs

    def get_option(self, argument):
        """Get an option to setup in a configuration file"""
        if argument in self.option_translator_dictionary:
            return self.option_translator_dictionary[argument]
        return argument  # maybe this should be None so it is obvious that there is an error

    def minimum_slab_size(self, device_name, default_to_2g=True):  # reused from stqe/host/vdo.py
        """Computing minimum slab size, used from stqe/host/vdo.py"""

        class VDODeviceNotFoundError(Exception):
            pass

        print(device_name)
        ret, device_size = run(cmd="lsblk | grep '%s ' " % device_name, return_output=True)
        print(device_name, ret, device_size)
        if ret != 0:
            raise VDODeviceNotFoundError()

        # ['└─myvg-vdotest', '252:2', '0', '4G', '0', 'lvm'][3] == '4G'
        size = device_size.split()[3]
        multipliers = ["M", "G", "T", "P", "E"]

        device_size = (float(size[:-1]) * (1024 ** multipliers.index(size[-1:]))).__int__()
        max_number_of_slabs = 8192
        minimum_size = max(
            2 ** int(device_size / max_number_of_slabs).bit_length(), 128
        )  # reused from stqe/host/vdo.py
        if default_to_2g and minimum_size < 2048:
            return "2G"
        return f"{str(minimum_size)}M"

    def check_conf_value(self, val):
        """Remove units K, G, T and convert accordingly"""
        if "fail" in self.tmp_kwargs["name"]:
            return val  # don't check

        units = {
            "K": 0.001,  # 1/1024 = 0.00097656,
            "k": 0.001,
            "M": 1,
            "m": 1,
            "G": 1024,  # 1000
            "g": 1024,
            "T": 1048576,  # 1000000
            "t": 1048576
            # "B": 1  # should do nothing
        }

        if re.match(r"[0-9]+\.[0-9]+", str(val)):  # float
            return val

        match = re.match(r"([0-9]+)(.+|)", str(val))
        if not match:
            return val

        multiplier = 1
        if len(match.groups()) == 2 and match.group(2):  # should not be empty
            try:
                multiplier = units[match.group(2)]
            except KeyError:  # beacause of "512B"
                print("Argument value: ", match.group(2), ".")
                print(f"Original {val}; will return {val[:-1]}")
                return val[:-1]
        return str(int(match.group(1)) * multiplier)

    def check_config_arguments(self, **kwargs):
        """Check arguments in config"""
        if "slab_size" in kwargs:
            device_name = kwargs["device"].split("/")[-1]
            # Slab_size value should be in MB
            kwargs["slab_size"] = self.check_conf_value(self.minimum_slab_size(device_name))

        return kwargs

    def create_profile_file(self, profile_name, args_to_remove, **kwargs):
        """category {
            arg1 = value,
            arg2 = value
        }
        """

        run(f"rm /etc/lvm/profile/{profile_name}.profile")

        # creates metadataprofile ONLY ONE CATEGORY NOW!!!!
        # update 28022022 don't know if it is true, after code review this should do multiple categories...
        to_write = {}  # category: [list of values]
        with open(f"/etc/lvm/profile/{profile_name}.profile", "w", encoding="utf-8") as fp:
            # create the structure of the document and save it in a string
            for arg, value in kwargs.items():
                if arg in self.config_arguments:  # do just what is necessary
                    category = self.get_category(arg)
                    if category not in to_write:
                        to_write[category] = []

                    to_write[category].append(f"{self.get_option(arg)}={self.check_conf_value(value)}")
                    args_to_remove.append(arg)

            str_to_write = ""
            for category in to_write:
                for value in to_write.values():
                    str_to_write += f"{category} {{\n"
                    for opt in value:
                        str_to_write += f"\t{opt}\n"
                str_to_write += "}\n"
            fp.write(str_to_write)  # in 'with' section there is no need for fp.close()

        return args_to_remove

    def replace_multiple_option_values(self, **kwargs):
        """Check and replace what is defined in .fmf test file with corresponding value from man page
        e.g. 'enabled' -> 'y'
        """
        for arg in self.multiple_option_arguments:
            if arg in kwargs:
                kwargs = self._set_kwargs_value(arg, **kwargs)
        # kwargs = {self._set_kwargs_value(i, **kwargs) for i in self.multiple_option_arguments if i in kwargs}
        return kwargs

    def remove_vdo_arguments(self, **kwargs):
        """The rest of VDO arguments need to be removed otherwise unsupported arguments
        will propagate into the command
        """
        arg_to_remove = [arg for arg in kwargs if arg in self.vdo_arguments]
        for arg in arg_to_remove:
            kwargs.pop(arg)
        return kwargs

    def get_category(self, argument):
        """Get the list that contains the argument I am looking for

        # not working for "check" (there are multiple) :( not fixed :/
        """

        for one_list in list(self.config_options.values()):
            if argument in one_list or self.option_translator_dictionary[argument] in one_list:
                arguments_of_category = one_list
                break  # find the first one

        # using the list containing the argument as item of the list
        index = list(self.config_options.values()).index(arguments_of_category)
        return list(self.config_options.keys())[index]

    def check(self, **kwargs):
        """Check arguments compatibility with command"""
        if "name" in kwargs:
            kwargs.pop("name")  # remove name, because it is not used in lvmvdo

        if "size" in kwargs:
            # value format for size parameter
            return bool(re.match(r"^([\+|\-]|)[0-9]+[T|G|M|K]$", kwargs["size"]))
        return True

    @staticmethod
    def remove_args(to_remove, **kwargs):
        """Return only what is intended"""

        return {k: v for k, v in kwargs.items() if k not in to_remove}

    @run_command
    def lvcreate(self, **kwargs):
        print("LVCREATE", kwargs)
        cmd = "lvcreate --vdo "

        # config part START
        args_to_remove = []
        use_conf_file = False

        if use_conf_file:
            profile_name = "vdo_create"
            args_to_remove = self.create_profile_file(profile_name, args_to_remove, **kwargs)
            cmd += f"--metadataprofile {profile_name} "
            print_profile_file(profile_name)
        elif self.number_of_conf_args(**kwargs) != 0:
            cmd += "--config"
            # --config="    for condition == 1
            # --config '   else
            cmd += '="' if self.number_of_conf_args(**kwargs) == 1 else " '"

            for arg, value in kwargs.items():
                if arg in self.config_arguments:  # do just what is necessary
                    if self.number_of_conf_args(**kwargs) == 1:
                        # --config="cat/opt=val
                        cmd += f"{self.get_category(arg)}/{self.get_option(arg)}={self.check_conf_value(value)}"
                    else:
                        # --config 'cat/opt=val cat/opt=val cat/opt=val
                        cmd += f"{self.get_category(arg)}/{self.get_option(arg)}={self.check_conf_value(value)} "
                    args_to_remove.append(arg)

            # --config="cat/opt=val"                          == 1
            # --config 'cat/opt=val cat/opt=val cat/opt=val'  else
            cmd += '" ' if self.number_of_conf_args(**kwargs) == 1 else "' "

        # some arguments are used as config and are no longer needed
        kwargs = self.remove_args(args_to_remove, **kwargs)
        # config part END

        kwargs["yes"] = "-y"  # see lvremove

        return cmd, kwargs

    @run_command
    def lvremove(self, **kwargs):
        cmd = "lvremove "

        # lvmvdo uses human interface for some activities like removing vdo pools
        # in automation we don't want this, so we add '-y' to force it
        kwargs["yes"] = "-y"
        return cmd, kwargs

    @run_command
    def lvconvert(self, **kwargs):
        cmd = "lvconvert --vdo "
        return cmd, kwargs

    @run_command
    def lvchange(self, **kwargs):
        cmd = "lvchange "
        if "vdo_name" in kwargs:
            kwargs.pop("vdo_name")
        return cmd, kwargs

    @run_command
    def lvextend(self, **kwargs):
        cmd = "lvextend "
        return cmd, kwargs

    @run_command
    def lvmconfig(self, **kwargs):
        cmd = "lvmconfig "
        return cmd, kwargs



#!/usr/bin/python

import os

import pytest
from libsan.host.cmdline import run


def test_unicode():
    # Read unicode text file from same directory the script is located.
    test_dir = os.path.dirname(__file__)
    if run("cat %s/unicode.txt" % test_dir) != 0:
        pytest.fail("Unable to read unicode.txt")



# Copyright (C) 2018 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""targetd.py: Module to manipulate functionality provided by targetd."""


__author__ = "Filip Suba"
__copyright__ = "Copyright (c) 2018 Red Hat, Inc. All rights reserved."

import os

from libsan import _print
from libsan.host import linux
from libsan.host.cmdline import run

targetd_cfg = "/etc/target/targetd.yaml"
targetd_directory = "/etc/target/"


def ssl_change(enable=False):
    """Changes variable ssl in targetd config to "false".
    :return:
    True: Variable changed successfully.
    False: Targetd config does not exist.
    """
    change = False
    if not os.path.isfile(targetd_cfg):
        _print("FAIL: Targetd config does not exist.")
        return False
    with open(targetd_cfg, encoding="UTF-8") as config:
        lines = config.readlines()
    for line in lines:
        if line.startswith("ssl:") and str(enable).lower() not in line:
            change = True
            lines[lines.index(line)] = "ssl: %s\n" % str(enable).lower()
            break

    if change:
        with open(targetd_cfg, "w", encoding="UTF-8") as config:
            config.writelines(lines)
    return True


def set_password(password=None):
    """Function for setting password in targetd config.
    :return:
    True: Password changed successfully.
    False: Targetd config does not exist.
    """
    pass_line = "password:\n"

    if not os.path.isfile(targetd_cfg):
        _print("FAIL: Targetd config does not exist.")
        return False
    with open(targetd_cfg, encoding="UTF-8") as config:
        lines = config.readlines()
    if password:
        pass_line = "password: %s\n" % password
    for line in lines:
        if line.startswith("password:"):
            lines[lines.index(line)] = pass_line
            break

    with open(targetd_cfg, "w", encoding="UTF-8") as config:
        config.writelines(lines)
    return True


def remove_file(name):
    """Removes a file from targetd directory.
    :param name: Name of a file in targetd directory
    :return:
    True: File removed successfully.
    False: Could not find SSL key in targetd directory or Could not remove SSL key.
    """
    if not os.path.isfile(targetd_directory + name):
        _print("FAIL: Could not find SSL key in /etc/target")
        return False
    cmd = f"rm -f {targetd_directory}{name}"
    ret = run(cmd)
    if ret != 0:
        _print("FAIL: Could not remove SSL key: %s" % name)
        return False
    return True


def gen_ssl_key(name):
    """Generates new private key and exports it as ".pem" file.
    :param name: Name for the export file.
    :return:
    True: Private key exported successfully.
    False: Could not generate SSL key or Could not change permissions.
    """

    print("Changing working directory to %s" % targetd_directory)
    pwd = os.path.abspath(os.curdir)
    os.chdir(targetd_directory)
    if os.path.abspath(os.curdir) != targetd_directory[:-1]:
        _print("FAIL: Could not change working directory to %s" % targetd_directory)
        return False

    cmd = "openssl genrsa -out %s.pem 2048" % (targetd_directory + name)
    ret = run(cmd)
    if ret != 0:
        _print("FAIL: Could not generate SSL key with name %s" % name)
        return False

    print("Changing working directory to %s" % pwd)
    os.chdir(pwd)
    if os.path.abspath(os.curdir) != pwd:
        _print("FAIL: Could not change working directory to %s" % pwd)
        return False

    cmd = f"chmod 600 {targetd_directory}{name}.pem"
    ret = run(cmd)
    if ret != 0:
        _print("FAIL: Could not change permissions to 600")
        return False
    return True


def gen_ssl_cert(name, key_name):
    """Generates certificate and exports it as ".pem" file. Function for testing only!
    :param name: Name for the export file.
    :param key_name: Name of a private key to use.
    :return:
    True: Certificate exported successfully.
    False: Provided key does not exist.
           Could not change permissions.
           Could not change working directory.
    """

    key = targetd_directory + key_name
    cert = targetd_directory + name

    print("Changing working directory to %s" % targetd_directory)
    pwd = os.path.abspath(os.curdir)
    os.chdir(targetd_directory)
    if os.path.abspath(os.curdir) != targetd_directory[:-1]:  # removes slash at the end of targetd_directory path
        _print("FAIL: Could not change working directory to %s" % targetd_directory)
        return False

    cmd = f"openssl req -new -x509 -subj /CN=CZ -key {key}.pem -out {cert}.pem -days 9999"
    ret = run(cmd)
    if ret != 0:
        _print(f"FAIL: Could not generate SSL cert with name {cert} and SSL key {key}")
        return False

    print("Changing working directory to %s" % pwd)
    os.chdir(pwd)
    if os.path.abspath(os.curdir) != pwd:
        _print("FAIL: Could not change working directory to %s" % pwd)
        return False

    cmd = "chmod 600 %s.pem" % cert
    ret = run(cmd)
    if ret != 0:
        _print("FAIL: Could not change permissions to 600")
        return False

    return True


def tls_status():
    """Checking if TLS is enabled.
    :return:
    True: TLS is enabled.
    False: TLS is disabled.
    """
    cmd = 'systemctl status targetd | grep "TLS yes"'
    retcode = run(cmd)
    if retcode != 0:
        _print("FAIL: TLS is not enabled")
        return False
    return True


def targetd_status(retcode=0):
    """Checks targetd service status.
    :param retcode: Expected return code.
    :return:
    True: Expected return code is equal to actual return code.
    False: Return code is different than expected.
    """
    ret = linux.service_status("targetd")
    if retcode != ret:
        _print("FAIL: Return code is different than expected!")
        print(f"Expected ret: {retcode}, got: {ret}")
        return False
    return True



#!/usr/bin/python

# Copyright (C) 2021 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


import unittest
from unittest.mock import patch

from libsan.host import net
from libsan.host.linux import is_docker

MAC = "7E:CD:60:1E:AC:6E"
NOTMAC = "7E:CD:60:1E:AC"
BADMAC = "7ecd.601e.ac6e"
LOWMAC = "7e:cd:60:1e:ac:6e"
NOTVALID = "ThisIsNotValid"
LOCALHOST = "127.0.0.1"
NETMASK24 = "255.255.255.0"
CIDR24 = "24"
NOTIP = "256.256.256.256"
IPV6 = "fd6f:60d2:2d9:0:b7e:b6e6:7bf5:c17e"


class TestNet(unittest.TestCase):
    def test_is_mac(self):
        assert net.is_mac(MAC)
        assert not net.is_mac(NOTMAC)

    def test_get_nics(self):
        assert net.get_nics() is not None

    def test_get_mac_of_nics(self):
        assert net.get_mac_of_nic(NOTVALID) is None
        assert net.get_mac_of_nic(net.get_nics()[-1]) is not None

    def test_get_nic_of_mac(self):
        assert net.get_nic_of_mac(MAC) is None
        assert net.get_nic_of_mac(net.get_mac_of_nic(net.get_nics()[-1])) is not None

    def test_get_ip_address_of_nic(self):
        assert net.get_ip_address_of_nic(NOTVALID) is None
        assert net.get_ip_address_of_nic(net.get_nics()[-1]) is not None

    def test_get_ipv6_address_of_nic(self):
        assert net.get_ipv6_address_of_nic(NOTVALID) is None
        if not is_docker():
            assert net.get_ipv6_address_of_nic(net.get_nics()[-1]) is not None

    def test_get_nic_of_ip(self):
        assert net.get_nic_of_ip(NOTIP) is None
        assert net.get_nic_of_ip(LOCALHOST) is not None

    def test_driver_of_nic(self):
        assert net.driver_of_nic(NOTVALID) is None

    def test_get_ip_version(self):
        assert net.get_ip_version(LOCALHOST) == 4
        assert net.get_ip_version(IPV6) == 6
        assert net.get_ip_version(NOTIP) is None

    def test_standardize_mac(self):
        assert net.standardize_mac(BADMAC) == LOWMAC
        assert net.standardize_mac(NOTMAC) is None

    def test_convert_netmask(self):
        assert int(CIDR24) == net.convert_netmask(NETMASK24)
        assert int(CIDR24) == net.convert_netmask(CIDR24)
        assert net.convert_netmask(NOTVALID) is None

    def test_nm_get_conn(self):
        assert net.nm_get_conn(MAC) is None
        assert net.nm_get_conn(NOTVALID) is None

    def test_nm_get_conn_iface(self):
        assert net.nm_get_conn_iface(NOTVALID) is None

    def test_nm_get_conn_uuid(self):
        assert net.nm_get_conn_uuid(NOTVALID) is None

    def test_nm_get_conn_from_dev(self):
        assert net.nm_get_conn_from_dev(NOTVALID) is None

    def test_nm_get_dev_from_conn(self):
        assert net.nm_get_dev_from_conn(NOTVALID) is None

    def test_nm_conn_up(self):
        assert not net.nm_conn_up(NOTVALID)

    def test_nm_set_ip(self):
        assert not net.nm_set_ip(NOTVALID, NOTIP)
        assert not net.nm_set_ip(NOTVALID, IPV6)

    def test_nm_dev_mod_success(self):
        with patch("libsan.host.net.run") as run_func:
            run_func.return_value = [0, ""]
            assert net.nm_dev_mod("enp17s0f1", "connection.autoconnect", "yes")

    def test_nm_dev_mod_failure(self):
        with patch("libsan.host.net.run") as run_func:
            run_func.return_value = [1, "Mocked failure"]
            assert not net.nm_dev_mod("enp17s0f1", "connection.autoconnect", "yes")

    def test_nm_set_ip_success(self):
        with patch("libsan.host.net.run") as run_func:
            run_func.return_value = [0, ""]
            assert net.nm_set_ip("enp17s0f1", "192.168.10.18", activate=False)

    def test_nm_set_ip_failure(self):
        with patch("libsan.host.net.run") as run_func:
            run_func.return_value = [1, "Mocked failure"]
            assert not net.nm_set_ip("enp17s0f1", "192.168.10.18", activate=False)



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""size.py: Module to convert human size <=> bytes."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re  # regex

from libsan import _print

size_human_regex = re.compile(r"([\-0-9\.]+)(Ki|Mi|Gi|Ti|Ei|Zi){0,1}B$")


def size_human_check(size_human):
    size_human = str(size_human)
    m = size_human_regex.match(size_human)
    if not m:
        _print("FAIL: size_human_check() - incorrect number format %s" % size_human)
        return False
    return True


def size_human_2_size_bytes(size_human):
    """Usage
        size_human_2_size_bytes(size_human)
    Purpose
        Convert human readable stander size to B
    Parameter
        size_human     # like '1KiB'
    Returns
        size_bytes     # like 1024
    """
    if not size_human:
        return None

    # make sure size_human is a string, could be only numbers, for example
    size_human = str(size_human)
    if not re.search(r"\d", size_human):
        # Need at least 1 digit
        return None

    m = size_human_regex.match(size_human)
    if not m:
        if re.match(r"^\d+$", size_human):
            # Assume size is already in bytes
            return size_human
        _print("FAIL: '%s' is an invalid human size format" % size_human)
        return None

    fraction = 0
    # check if number is fractional
    f = re.match(r"(\d+)\.(\d+)", m.group(1))
    if f:
        number = int(f.group(1))
        fraction = int(f.group(2))
    else:
        number = int(m.group(1))

    unit = m.group(2)
    if not unit:
        unit = "B"

    for valid_unit in ["B", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if unit == valid_unit:
            if unit == "B":
                # cut any fraction if was given, as it is not valid
                return str(number)
            return int(number + fraction)
        number *= 1024
        fraction *= 1024
        fraction /= 10
    return int(number + fraction)


def size_bytes_2_size_human(num):
    if not num:
        return None

    # Even if we receive string we convert so we can process it
    num = int(num)
    for unit in ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB"]:
        if abs(num) < 1024.0:
            size_human = f"{num:3.1f}{unit}"
            # round it down removing decimal numbers
            size_human = re.sub(r"\.\d+", "", size_human)
            return size_human
        num /= 1024.0
    # Very big number!!
    size_human = "{:.1f}{}".format(num, "Yi")
    # round it down removing decimal numbers
    size_human = re.sub(r"\.\d+", "", size_human)
    return size_human



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""linux.py: Module to get information from Linux servers."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import errno
import os.path
import re  # regex
import signal
import subprocess
import sys
import time
from datetime import datetime
from os import listdir

import requests
from configobj import ConfigObj
from pkg_resources import parse_version

import libsan.host.mp
import libsan.host.scsi
from libsan import _print
from libsan.host.cmdline import exists, run


def hostname():
    ret, host = run("hostname", verbose=False, return_output=True)
    if ret != 0:
        _print("FAIL: hostname() - could not run command")
        print(host)
        return None
    return host


def linux_distribution():
    # Not using platform module, as it doesn't provide needed information
    # on recent python3 versions
    # see: https://bugzilla.redhat.com/show_bug.cgi?id=1920385
    # see: https://bugs.python.org/issue28167
    data = {}
    with open("/etc/os-release", encoding="UTF-8") as f:
        for line in f:
            line = line.strip()
            if line == "":
                continue
            k, v = line.split("=")
            data[k] = v.strip('"')
    return [data["ID"], data["VERSION_ID"]]


def dist_release():
    """Find out the release number of Linux distribution."""
    # We are base on output of lsb_release -r -s' which is shipped by redhat-lsb rpm.
    # ret, release = run("lsb_release --release --short", verbose=False, return_output=True)
    # if ret == 0:
    #    return release
    dist = linux_distribution()
    if not dist:
        _print("FAIL: Could not determine dist release!")
        return None
    return dist[1]


def dist_ver():
    """Check the Linux distribution version."""
    release = dist_release()
    if not release:
        return None
    m = re.match(r"(\d+).\d+", release)
    if m:
        return int(m.group(1))

    # See if it is only digits, in that case return it
    m = re.match(r"(\d+)", release)
    if m:
        return int(m.group(1))

    _print("FAIL: dist_ver() - Invalid release output %s" % release)
    return None


def dist_ver_minor():
    """Check the Linux distribution minor version.
    For example: RHEL-7.4 returns 4
    """
    release = dist_release()
    if not release:
        return None
    m = re.match(r"\d+.(\d+)", release)
    if m:
        return int(m.group(1))

    _print("FAIL: dist_ver_minor() - Release does not seem to have minor version: %s" % release)
    return None


def dist_name():
    """Find out the name of Linux distribution."""
    # We are base on output of lsb_release -r -s' which is shipped by redhat-lsb rpm.
    # ret, release = run("lsb_release --release --short", verbose=False, return_output=True)
    # if ret == 0:
    #    return release
    dist = linux_distribution()
    if not dist:
        _print("FAIL: dist_name() - Could not determine Linux dist name")
        return None

    if dist[0] == "rhel":
        return "RHEL"

    return dist[0]


def service_start(service_name):
    """Start service
    The arguments are:
    \tNone
    Returns:
    \tTrue: Service started
    \tFalse: There was some problem
    """
    cmd = "systemctl start %s" % service_name
    has_systemctl = True

    if not exists("systemctl"):
        has_systemctl = False
    if not has_systemctl:
        cmd = "service %s start" % service_name

    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not start %s" % service_name)
        if has_systemctl:
            run("systemctl status %s" % service_name)
            run("journalctl -xn")
        return False
    return True


def service_stop(service_name):
    """Stop service
    The arguments are:
    \tName of the service
    Returns:
    \tTrue: Service stopped
    \tFalse: There was some problem
    """
    cmd = "systemctl stop %s" % service_name
    has_systemctl = True

    if not exists("systemctl"):
        has_systemctl = False
    if not has_systemctl:
        cmd = "service %s stop" % service_name

    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not stop %s" % service_name)
        if has_systemctl:
            run("systemctl status %s" % service_name)
            run("journalctl -xn")
        return False
    return True


def service_restart(service_name):
    """Restart service
    The arguments are:
    \tName of the service
    Returns:
    \tTrue: Service restarted
    \tFalse: There was some problem
    """
    cmd = "systemctl restart %s" % service_name
    has_systemctl = True

    if not exists("systemctl"):
        has_systemctl = False
    if not has_systemctl:
        cmd = "service %s restart" % service_name
    service_timestamp = get_service_timestamp(service_name)
    if service_timestamp is not None:
        timestamp_struct = time.strptime(service_timestamp, "%a %Y-%m-%d %H:%M:%S %Z")
        actual_time = time.localtime()
        if time.mktime(actual_time) - time.mktime(timestamp_struct) < 5:
            print("Waiting 5 seconds before restart.")
            time.sleep(5)
    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not restart %s" % service_name)
        if has_systemctl:
            run("systemctl status %s" % service_name)
            run("journalctl -xn")
        return False
    return True


def service_enable(service_name, now=False):
    """Enable service
    The arguments are:
    \tName of the service
    Returns:
    \tTrue: Service got enabled
    \tFalse: There was some problem
    """
    cmd = "systemctl enable %s" % service_name
    has_systemctl = True

    if not exists("systemctl"):
        has_systemctl = False
    if not has_systemctl:
        cmd = "chkconfig %s stop" % service_name

    if now and has_systemctl:
        cmd = cmd + " --now"

    retcode = run(cmd, return_output=False, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not enable %s" % service_name)
        if has_systemctl:
            run("systemctl status %s" % service_name)
            run("journalctl -xn")
        return False
    return True


def service_status(service_name, verbose=True):
    """Check service status
    The arguments are:
    \tName of service
    Returns:
    \t 0 - service is running and OK
    \t 1 - service is dead and /run pid file exists
    \t 2 - service is dead and /lock lock file exists
    \t 3 - service is not running
    \t 4 - service could not be found.
    \t False - something went wrong.
    """
    cmd = "systemctl status %s" % service_name
    has_systemctl = True

    if not exists("systemctl"):
        has_systemctl = False
    if not has_systemctl:
        cmd = "service %s status" % service_name

    retcode = run(cmd, return_output=False, verbose=verbose)
    if retcode == 0:
        _print("INFO: Service %s is running." % service_name)
    elif retcode == 1:
        _print("INFO: Service %s is dead and /run pid file exists." % service_name)
    elif retcode == 2:
        _print("INFO: Service %s is dead and /lock lock file exists." % service_name)
    elif retcode == 3:
        _print("INFO: Service %s is not running." % service_name)
    elif retcode == 4:
        _print("INFO: Service %s could not be found." % service_name)
    else:
        _print(f"INFO: Service {service_name} returned unknown code {retcode}.")
    return retcode


def is_service_running(service_name):
    """Check if service is running
    The arguments are:
    \tName of service
    Returns:
    \t True: service is running
    \t False: service is not running
    """

    return service_status(service_name, verbose=False) == 0


def os_arch():
    ret, arch = run("uname -m", verbose=False, return_output=True)
    if ret != 0:
        _print("FAIL: could not get OS arch")
        return None

    return arch


def is_installed(pack, verbose=False):
    """Checks if package is installed"""
    ret, ver = run("rpm -q %s" % pack, verbose=False, return_output=True)
    if ret == 0:
        if verbose:
            print(f"INFO: {pack} is installed ({ver})")
        return True

    if verbose:
        print("INFO: %s is not installed " % pack)
    return False


def install_package(pack, check=True, verbose=True):
    """Install a package "pack" via `yum|dnf install -y`"""
    # Check if package is already installed
    if check and is_installed(pack, verbose):
        return True

    packmngr = "yum"
    if is_installed("dnf"):
        packmngr = "dnf"

    if run(f"{packmngr} install -y {pack}") != 0:
        msg = "FAIL: Could not install %s" % pack
        _print(msg)
        return False

    if verbose:
        print("INFO: %s was successfully installed" % pack)
    return True


def package_version(pkg):
    """Get the version of specific package"""
    if not pkg:
        _print("FAIL: package_version requires package name as parameter")
        return None

    release = dist_name().lower()
    if not release:
        _print("FAIL: package_version() - Couldn't get release")
        return None
    if exists("rpm"):
        ret, output = run("rpm -qa --qf='%%{version}.%%{release}' %s" % pkg, return_output=True, verbose=False)
        if ret != 0:
            _print("FAIL: Could not get version for package: %s" % pkg)
            print(output)
            return None
        return output

    _print("FAIL: package_version() - Unsupported release: %s" % release)
    return None


def compare_version(package, version, release, equal=True):
    """:param package: package name
    :param version: package version
    :param release: package release
    :param equal: return value if packaged are equal version (default 'True')
    :return: Returns 'True' if installed version is newer or equal than asked, 'False' if older
             Return 'None' if the package is not installed (not found)
    """
    pkg = package_version(package)
    if pkg is None:
        return None
    if any([True for x in pkg if not isinstance(x, int)]):
        # cut ending of package name containing string, parse_version() does not work with ints and strings at once
        pack = []
        for p in pkg[:].split("."):
            try:
                int(p)
            except ValueError:
                break
            pack.append(p)
        pkg = ".".join(pack)
    if equal:
        return parse_version(version + "." + release) <= parse_version(pkg)
    return parse_version(version + "." + release) < parse_version(pkg)


def wait_udev(sleeptime=15):
    """Wait udev to finish. Often used after scsi rescan."""
    _print("INFO: Waiting udev to finish storage scan")
    # For example, on RHEL 7 scsi_wait_scan module is deprecated
    if run("modinfo scsi_wait_scan", verbose=False) == 0:
        run("modprobe -q scsi_wait_scan")
        run("modprobe -r -q scsi_wait_scan")

    run("udevadm settle")
    sleep(sleeptime)

    return True


def get_all_loaded_modules():
    """Check /proc/modules and return a list of all modules that are loaded"""
    cmd = 'cat /proc/modules | cut -d " " -f 1'
    ret, output = run(cmd, return_output=True, verbose=False)
    if ret != 0:
        _print("FAIL: load_module() - Could not execute: %s" % cmd)
        print(output)
        return None

    modules = output.split("\n")
    return modules


def load_module(module):
    """Run modprobe using module with parameters given as input
    Parameters:
    \tmodule:       module name and it's parameters
    """
    if not module:
        _print("FAIL: load_module() - requires module parameter")
        return False
    cmd = "modprobe %s" % module
    if run(cmd) != 0:
        _print("FAIL: load_module() - Could not execute: %s" % module)
        return False
    return True


def unload_module(module_name, remove_dependent=False):
    """Run rmmod to unload module
    Parameters:
    \tmodule_name:       module name
    """
    if not module_name:
        _print("FAIL: unload_module() - requires module_name parameter")
        return False
    cmd = f"modprobe -r {module_name}"

    if remove_dependent:
        dep_modules = get_dependent_modules(module_name)
        if dep_modules:  # print info only if there are any modules to remove
            _print(f"INFO: Removing modules dependent on {module_name}")
            for module in dep_modules:
                if not unload_module(module, remove_dependent=remove_dependent):
                    _print("FAIL: unload_module() - Could not unload dependent modules")
                    return False

    if run(cmd) != 0:
        _print(f"FAIL: unload_module() - Could not unload: {module_name}")
        return False

    return True


def get_dependent_modules(module_name):
    """Returns list of modules that loaded this module as a dependency
    Useful when removing parent modules (error "Module is in use by: ")
    Parameters:
    \tmodule_name:      module_name
    """
    if not module_name:
        _print("FAIL: get_dependent_modules() - requires module_name parameter")
        return None
    cmd = f'cat /proc/modules | grep -Ew "^{module_name}" | cut -d \' \' -f 4 | tr "," " "'
    ret, dependent_modules = run(cmd, return_output=True, verbose=False)
    if dependent_modules == "-":
        return []  # No dependent modules found
    if ret != 0:
        _print("FAIL: get_dependent_modules() - failed to get a list of modules")
        return None
    return dependent_modules.split()


def is_module_loaded(module_name):
    """Check if given module is loaded
    Parameters:
    \tmodule_name:      module_name
    """
    if module_name in get_all_loaded_modules():
        return True
    return False


def sleep(duration):
    """It basically call sys.sleep, but as stdout and stderr can be buffered
    We flush them before sleep
    """
    sys.stdout.flush()
    sys.stderr.flush()
    time.sleep(duration)
    return


def is_mounted(device=None, mountpoint=None):
    """Check if mountpoint is already mounted"""
    if device and run("mount | grep %s" % device, verbose=False) != 0:
        return False
    if mountpoint and run("mount | grep %s" % mountpoint, verbose=False) != 0:
        return False
    return True


def mount(device=None, mountpoint=None, fs=None, options=None):
    cmd = "mount"
    if fs:
        cmd += " -t %s" % fs
    if options:
        cmd += " -o %s" % options
    if device:
        cmd += " %s" % device
    if mountpoint:
        cmd += " %s" % mountpoint
    if run(cmd) != 0:
        _print("FAIL: Could not mount partition")
        return False

    return True


def umount(device=None, mountpoint=None):
    cmd = "umount"
    if device:
        cmd += " %s" % device
        if not is_mounted(device):
            # Device is not mounted
            return True

    if mountpoint:
        cmd += " %s" % mountpoint
        if not is_mounted(mountpoint=mountpoint):
            # Device is not mounted
            return True

    if run(cmd) != 0:
        _print("FAIL: Could not umount partition")
        return False

    return True


def get_default_fs():
    """Return the default FileSystem for this release"""
    if dist_name() == "RHEL" and dist_ver() > 6:
        return "xfs"

    return "ext4"


def run_cmd_background(cmd):
    """Run Command on background
    Returns:
        subprocess.
            PID is on process.pid
            Exit code is on process.returncode (after run process.communicate())
        Wait for process to finish
            while process.poll() is None:
                linux.sleep(1)
        Get stdout and stderr
            (stdout, stderr) = process.communicate()
    """
    process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    if not process:
        _print("FAIL: Could not run '%s' on background" % cmd)
        return None
    _print("INFO: running %s on background. PID is %d" % (cmd, process.pid))
    return process


def kill_pid(pid):
    os.kill(pid, signal.SIGTERM)
    sleep(1)
    if check_pid(pid):
        os.kill(pid, signal.SIGKILL)
        sleep(1)
        if check_pid(pid):
            return False
    return True


def kill_all(process_name):
    ret = run("killall %s" % process_name, verbose=None)
    # Wait few seconds for process to finish
    sleep(3)
    return ret


def check_pid(pid):
    """Check there is a process running with this PID"""
    # try:
    # #0 is the signal, it does not kill the process
    # os.kill(int(pid), 0)
    # except OSError:
    # return False
    # else:
    # return True
    try:
        return os.waitpid(pid, os.WNOHANG) == (0, 0)
    except OSError as e:
        if e.errno != errno.ECHILD:
            raise


def time_stamp(in_seconds=False):
    now = datetime.now()

    # ts = "%s%s%s%s%s%s" % (now.year, now.month, now.day, now.hour, now.minute, now.second)
    ts = now.strftime("%Y%m%d%H%M%S")
    if in_seconds:
        ts = now.strftime("%s")
    return ts


def kernel_command_line():
    """Return the kernel command line used to boot"""
    retcode, output = run("cat /proc/cmdline", return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: could not get kernel command line")
        print(output)
        return None
    return output


def kernel_version():
    """Usage
        kernel_version()
    Purpose
        Check out running kernel version. The same as output of `uname -r`
    Parameter
        N/A
    Returns
        kernel_version
    """
    retcode, output = run("uname -r", return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: could not get kernel version")
        print(output)
        return None
    # remove arch detail and kernel type
    output = re.sub(r"\.%s.*" % os_arch(), "", output)
    return output


def kernel_type():
    """Usage
        kernel_type()
    Purpose
        Check the kernel type. Current we support detection of these types:
            1. default kernel.
            2. debug kernel.
            3. rt kernel.
    Parameter
        N/A
    Returns
        kernel_type        # 'debug|rt|default'
    """
    retcode, version = run("uname -r", return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: kernel_type() - could not get kernel version")
        print(version)
        return None

    if re.match(r".*\.debug$", version):
        return "debug"

    if re.match(r".*\.rt$", version):
        return "rt"

    return "default"


def kmem_leak_start():
    """Usage
        kmem_leak_start()
    Purpose
        Start and clear kernel memory leak detection.
    Parameter
        N/A
    Returns
        True
          or
        False       # not debug kernel or failure found
    """
    k_type = kernel_type()

    if not k_type or k_type != "debug":
        _print("WARN: Not debug kernel, will not enable kernel memory leak check")
        return False

    arch = os_arch()
    if arch == "i386" or arch == "i686":
        print("INFO: Not enabling kmemleak on 32 bits server.")
        return False

    k_commandline = kernel_command_line()
    if not re.search("kmemleak=on", k_commandline):
        _print("WARN: kmem_leak_start(): need 'kmemleak=on' kernel_option to enable kernel memory leak detection")

    check_debugfs_mount_cmd = 'mount | grep "/sys/kernel/debug type debugfs"'
    retcode = run(check_debugfs_mount_cmd, verbose=False)
    if retcode != 0:
        # debugfs is not mounted
        mount_debugfs_cli_cmd = "mount -t debugfs nodev /sys/kernel/debug"
        run(mount_debugfs_cli_cmd, verbose=True)
        check_debugfs_mount_cmd = 'mount | grep "/sys/kernel/debug type debugfs"'
        retcode, output = run(check_debugfs_mount_cmd, return_output=True, verbose=False)
        if retcode != 0:
            _print("WARN: Failed to mount debugfs to /sys/kernel/debug")
            print(output)
            return False

    # enable kmemleak and clear
    _print("INFO: Begin kernel memory leak check")
    if run("echo scan=on > /sys/kernel/debug/kmemleak") != 0:
        return False
    if run("echo stack=on > /sys/kernel/debug/kmemleak") != 0:
        return False
    if run("echo clear > /sys/kernel/debug/kmemleak") != 0:
        return False
    return True


def kmem_leak_check():
    """Usage
        kmem_leak_check()
    Purpose
        Read out kernel memory leak check log and then clear it up.
    Parameter
        N/A
    Returns
        kmemleak_log
          or
        None       # when file '/sys/kernel/debug/kmemleak' not exists
                  # or no leak found
    """
    sysfs_kmemleak = "/sys/kernel/debug/kmemleak"
    if not os.path.isfile(sysfs_kmemleak):
        return None

    with open(sysfs_kmemleak) as f:
        if not f:
            _print("FAIL: Could not read %s" % sysfs_kmemleak)
            return None
        kmemleak_log = f.read()

    if kmemleak_log:
        _print("WARN: Found kernel memory leak:\n%s" % kmemleak_log)
        _print("INFO: Clearing memory leak for next check")
        run("echo 'clear' > %s" % sysfs_kmemleak)
        return kmemleak_log

    _print("INFO: No kernel memory leak found")
    return None


def kmem_leak_disable():
    """Usage
        kmem_leak_disable()
    Purpose
        Disable kmemleak by 'scan=off' and 'stack=off' to
        '/sys/kernel/debug/kmemleak'.
    Parameter
        N/A
    Returns
        True           # disabled or not enabled yet
          or
        False       # failed to run 'echo' command
    """
    sysfs_kmemleak = "/sys/kernel/debug/kmemleak"
    if not os.path.isfile(sysfs_kmemleak):
        return True

    _print("INFO: kmem_leak_disable(): Disabling kernel memory leak detection")
    ok1, ok1_output = run("echo scan=off > %s" % sysfs_kmemleak, return_output=True)
    ok2, ok2_output = run("echo stack=off > %s" % sysfs_kmemleak, return_output=True)
    if ok1 != 0 or ok2 != 0:
        print("FAIL: kmem_leak_disable(): Failed to disable kernel memory leak detection")
        print(ok1_output)
        print(ok2_output)
        return False

    _print("INFO: kmem_leak_disable(): Kernel memory leak detection disabled")
    return True


def query_os_info():
    """query_os_info()
    Purpose
        Query OS informations and set a reference below:
        os_info = {
            dist_name       = dist_name,
            dist_release    = dist_release,
            kernel_version  = kernel_version,
            os_arch         = os_arch,
            arch            = os_arch,
            pkg_arch        = pkg_arch, #as, for example, rpm on i386 could return different arch then uname -m

        }
    Parameter
        N/A
    Returns
        os_info_dict
            or
        None       # got error
    """

    os_info_dict = {
        "dist_name": dist_name(),
        "dist_version": dist_ver(),
        "dist_release": dist_release(),
        "os_arch": os_arch(),
        "arch": os_arch(),
        "pkg_arch": os_arch(),
        "kernel_version": kernel_version(),
        "kernel_type": kernel_type(),
    }
    return os_info_dict


def get_driver_info(driver):
    if not driver:
        _print("FAIL: get_driver_info() - requires driver parameter")
        return None

    sys_fs_path = "/sys/module"
    if not os.path.isdir(sys_fs_path):
        _print("FAIL: get_driver_info() - %s is not a valid directory" % sys_fs_path)
        return None

    sysfs_driver_folder = f"{sys_fs_path}/{driver}"
    if not os.path.isdir(sysfs_driver_folder):
        _print("FAIL: get_driver_info() - module %s is not loaded" % driver)
        return None

    driver_info = {}
    infos = ["srcversion", "version", "taint"]
    for info in infos:
        info_path = f"{sysfs_driver_folder}/{info}"
        if not os.path.isfile(info_path):
            continue
        _, output = run(f"cat {sysfs_driver_folder}/{info}", return_output=True, verbose=False)
        driver_info[info] = output

    sys_driver_parameter = "%s/parameters/" % sysfs_driver_folder
    if os.path.isdir(sys_driver_parameter):
        # Need to add driver parameters
        param_files = [
            f for f in listdir(sys_driver_parameter) if os.path.isfile(os.path.join(sys_driver_parameter, f))
        ]
        for param in param_files:
            _, output = run(f"cat {sys_driver_parameter}/{param}", return_output=True, verbose=False)
            if "parameters" not in driver_info:
                driver_info["parameters"] = {}
            driver_info["parameters"][param] = output
    return driver_info


def mkdir(new_dir):
    if os.path.isdir(new_dir):
        _print("INFO: %s already exist" % new_dir)
        return True
    cmd = "mkdir -p %s" % new_dir
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: could create directory %s" % new_dir)
        print(output)
        return False
    return True


def rmdir(dir_name):
    """Remove directory and all content from it"""
    if not os.path.isdir(dir_name):
        _print("INFO: %s does not exist" % dir_name)
        return True
    cmd = "rm -rf %s" % dir_name
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: could remove directory %s" % dir_name)
        print(output)
        return False
    return True


def mkfs(device_name, fs_type, force=False):
    """Create a Filesystem on device"""
    if not device_name or not fs_type:
        _print("INFO: mkfs() requires device_name and fs_type")
        return False

    force_option = "-F"
    if fs_type == "xfs":
        force_option = "-f"

    cmd = "mkfs.%s " % fs_type
    if force:
        cmd += "%s " % force_option
    cmd += device_name
    retcode, output = run(cmd, return_output=True, verbose=True)
    if retcode != 0:
        _print(f"FAIL: could create filesystem {fs_type} on {device_name}")
        print(output)
        return False
    return True


def sync(directory=None):
    cmd = "sync"
    if directory:
        cmd += " %s" % directory
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: could not sync")
        print(output)
        return False
    return True


def get_free_space(path):
    """Get free space of a path.
    Path could be:
    \t/dev/sda
    \t/root
    \t./
    """
    if not path:
        return None

    cmd = "df -B 1 %s" % path
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: get_free_space() - could not run %s" % cmd)
        print(output)
        return None
    fs_list = output.split("\n")
    # delete the header info
    del fs_list[0]

    if len(fs_list) > 1:
        # Could be the information was too long and splited in lines
        tmp_info = "".join(fs_list)
        fs_list[0] = tmp_info

    # expected order
    # Filesystem    1B-blocks       Used   Available Use% Mounted on
    free_space_regex = re.compile(r"\S+\s+\d+\s+\d+\s+(\d+)")
    m = free_space_regex.search(fs_list[0])
    if m:
        return int(m.group(1))
    return None


def get_block_device_name(device):
    """Returns kernel name from block device
    eg. lvm1 from /dev/mapper/lvm1
    """
    if not device.startswith("/dev/"):
        device = get_full_path(device)
    if not device:
        _print("FAIL: get_block_device_name - unknown device")
    cmd = "lsblk -ndlo NAME %s" % device
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: run %s" % cmd)
        print(output)
        return None
    return output


def get_full_path(device_name):
    """Returns full block device path, eg. from device: /dev/mapper/device"""
    cmds = [
        "lsblk -pnalo NAME  | grep %s -m1" % device_name,  # should be more robust
        "find /dev/ -name %s" % device_name,
    ]  # older OS(rhel-6), will fail with partitions

    for cmd in cmds:
        retcode, output = run(cmd, return_output=True, verbose=False)
        if retcode == 0 and output != "":
            return output

    _print("FAIL: get_full_path() - %s" % device_name)
    return None


def get_parent_device(child_device, only_direct=False):
    """Returns block device's parent device: eg. sda, nvme0n1
    child_device: eg. /dev/sda2, nvme0n1p1, /dev/mapper/device
    only_direct: returns only the direct parent. eg. lvm -> sda3, not sda
    """
    if not child_device.startswith("/dev/"):
        child_device = get_full_path(child_device)
    if not child_device:  # get_full_path would return None if device does not exist
        _print("FAIL: get_parent_device - unknown child_device '%s'" % child_device)
    cmd = "lsblk -nsl %s -o KNAME | tail -n 1" % child_device
    if only_direct:
        cmd = "lsblk -nsl %s -o KNAME | sed -n 2p" % child_device  # if no parent, returns nothing
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: run %s" % cmd)
        print(output)
        return None
    if output == "" or output == child_device:
        _print("WARN: get_parent_device - device has no parent")
        return None
    return output


def get_udev_property(device_name, property_key):
    """Given an /dev device name, returns specified property using udevadm
    :param device_name: eg. 'sda', 'mpatha', 'dm-0', 'nvme0n1', 'sr0', ...
    :param property_key: eg. 'ID_SERIAL', 'DM_WWN', 'ID_PATH', ...
    :return property_value: eg. for ID_SERIAL: '360fff19abdd9f5fb943525d45126ca27'
    """
    if not device_name:
        _print("WARN: get_udev_property() - requires device_name parameter")
        return None

    # Converts for example mpatha to /dev/mapper/mpatha or sda to /dev/sda
    device = libsan.host.linux.get_full_path(device_name)
    if not device:
        _print("FAIL: get_udev_property - unknown device_name '%s'" % device_name)

    # Trying to catch wrong key name when dm-multipath is used.
    if libsan.host.mp.is_mpath_device(device_name, print_fail=False):  # noqa: SIM102
        if property_key.startswith("ID_") and not property_key.startswith("ID_FS_"):
            property_key = property_key.replace("ID_", "DM_")

    ret, property_value = run(
        f"udevadm info -q property --name={device} | grep {property_key}= | cut -d = -f 2",
        return_output=True,
        verbose=False,
    )
    if ret:
        _print("WARN: Could not get udevadm info of device '%s'" % device)
        return None
    if not property_value:
        _print(f"WARN: Could not find property '{property_key}' in udevadm info of device '{device}'")
        return None

    return property_value


def get_boot_device(parent_device=False, full_path=False):
    """Returns boot device, eg. 'sda1'
    parent_device, eg. 'sda'
    full_path, eg. '/dev/sda1'
    """
    boot_mount = "/boot"
    root_mount = "/"
    # get boot device
    cmd = "mount | grep ' %s ' | cut -d ' ' -f 1" % boot_mount
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: run %s" % cmd)
        print(output)
        return None
    boot_device = output
    # get root device
    cmd = "mount | grep ' %s ' | cut -d ' ' -f 1" % root_mount
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: run %s" % cmd)
        print(output)
        return None
    root_device = output

    if not boot_device and not root_device:
        _print("FAIL: Could not find '/boot' and '/' mounted!")
        return None
    if not boot_device:
        # /boot is not mounted on openstack virtual machines
        _print("INFO: Could not find /boot mounted... Assuming this is a virtual machine")
        boot_device = root_device
    if boot_device == "overlay":
        _print("INFO: / mounted on overlay device. Assuming running in a container")
        return None

    if parent_device:
        boot_device = get_parent_device(boot_device)
    if full_path:
        return get_full_path(boot_device)
    return get_block_device_name(boot_device)


def is_dm_device(device_name):
    """Checks if device is mapped by device-mapper.
    :param device_name: eg. 'sda', 'mpatha', ...
    :return: True/False
    """
    # Converts for example mpatha to /dev/mapper/mpatha or sda to /dev/sda
    device = libsan.host.linux.get_full_path(device_name)
    if not device:
        _print("FAIL: is_dm_device - unknown device_name '%s'" % device_name)
        return False
    ret, name = run("udevadm info -q name --name=%s" % device, return_output=True)
    if ret:
        _print("FAIL: Could not get udevadm info for device '%s'" % device)
        return False
    if not name:
        _print("FAIL: Could not find udev name for '%s'" % device)
        return False

    if name.startswith("dm"):
        return True

    return False


def is_nvme_device(device):
    """Checks if device is nvme device."""
    return bool(re.match("^nvme[0-9]n[0-9]$", device))


def get_wwid_of_nvme(device):
    """Reads WWID from udev ID_WWN."""
    return get_udev_property(device, property_key="ID_WWN")


def get_device_wwid(device):
    """Given a SCSI, NVMe or multipath device, returns its WWID"""

    if device.startswith("vd"):
        print("INFO: %s: Presuming virtual disk does not have wwid." % device)
        return None

    serial = get_udev_property(device_name=device, property_key="ID_SERIAL")
    if not serial and is_dm_device(device):  # RHEL-6 workaround
        dm_uuid = get_udev_property(device_name=device, property_key="DM_UUID")
        serial = dm_uuid.replace("mpath-", "")
    if not serial:
        _print("INFO: get_device_wwid() - Could not find WWID for %s" % device)
        return None

    return serial


def remove_device_wwid(wwid):
    if not wwid:
        _print("FAIL: remove_device_wwid() - requires wwid as parameter")
        return False

    mpath_wwid = libsan.host.mp.mpath_name_of_wwid(wwid)
    if mpath_wwid:
        libsan.host.mp.remove_mpath(mpath_wwid)

    scsi_ids_wwid = libsan.host.scsi.scsi_ids_of_wwid(wwid)
    if scsi_ids_wwid:
        for scsi_id in scsi_ids_wwid:
            scsi_name = libsan.host.scsi.get_scsi_disk_name(scsi_id)
            if not scsi_name:
                continue
            _print("INFO: detaching SCSI disk %s" % scsi_name)
            libsan.host.scsi.delete_disk(scsi_name)
    return True


def clear_dmesg():
    cmd = "dmesg --clear"
    if dist_ver() < 7:
        cmd = "dmesg -c"
    run(cmd, verbose=False)
    return True


def get_regex_pci_id():
    regex_pci_id = r"(?:([0-0a-f]{4}):){0,1}"  # domain id (optional)
    regex_pci_id += r"([0-9a-f]{2})"  # bus id
    regex_pci_id += r":"
    regex_pci_id += r"([0-9a-f]{2})"  # slot id
    regex_pci_id += r"\."
    regex_pci_id += r"(\d+)"  # function id
    return regex_pci_id


def get_partitions(device):
    """Return a list of all parition numbers from the device"""
    if not device:
        _print("WARN: get_partitions() - requires device as parameter")
        return None

    cmd = "parted -s %s print" % device
    ret, output = run(cmd, verbose=False, return_output=True)
    if ret != 0:
        # _print("FAIL: get_partitions() - Could not read partition information from %s" % device)
        # print output
        return None

    lines = output.split("\n")
    if not lines:
        return None

    header_regex = re.compile(r"Number  Start   End     Size    Type")
    partition_regex = re.compile(r"\s(\d+)\s+\S+")
    partitions = []
    found_header = False
    for line in lines:
        if header_regex.match(line):
            found_header = True
            continue
        if found_header:
            m = partition_regex.match(line)
            if m:
                partitions.append(m.group(1))

    return partitions


def delete_partition(device, partition):
    """Delete specific partition from the device"""
    if not device or not partition:
        _print("FAIL: delete_partition() - requires device and partition as argument")
        return False

    cmd = f"parted -s {device} rm {partition}"
    ret, output = run(cmd, verbose=False, return_output=True)
    if ret != 0:
        _print("FAIL: delete_partition() - Could not delete partition %d from %s" % (partition, device))
        print(output)
        return False

    return True


def add_repo(name, address, metalink=False):
    """Adds yum repository to /etc/yum.repos.d/NAME.repo"""
    repo = "/etc/yum.repos.d/%s.repo" % name.lower()
    if os.path.isfile(repo):
        _print("INFO: Repo %s already exists." % repo)
        return True

    url = "baseurl=%s\n" % address
    if metalink:
        url = "metalink=%s\n" % address

    with open(repo, "w", encoding="UTF-8") as _file:
        _file.write(
            "[%s]\n" % name + "name=%s\n" % name + url + "enabled=1\n" + "gpgcheck=0\n" + "skip_if_unavailable=1\n"
        )

    return True


def download_repo_file(url, name=None, overwrite=True):
    """Downloads .repo file to /etc.repos.d/"""
    if not url:
        print("FAIL: repo file url argument required")
        return False
    if not name:
        name = url.split("/")[-1]
    if name[-5:] != ".repo":
        name = "{}{}".format(name, ".repo")
    path = "/etc/yum.repos.d/%s" % name

    if os.path.isfile(path):
        if overwrite is False:
            print("WARN: %s exits, skipping repo file download" % name)
            return True
        else:
            print("WARN: %s exits, overwriting .repo file" % name)
    try:
        r = requests.get(url, timeout=10)
    except requests.exceptions.RequestException as e:
        print("FAIL: Unable to reach repo url: %s" % e)
        return False
    try:
        with open(path, "wb") as f:
            f.write(r.content)
    except OSError:
        print("FAIL: Unable to write %s file" % path)
        return False

    return True


def del_repo(name):
    """Removes .repo file"""
    repo = "/etc/yum.repos.d/%s.repo" % name.lower()
    if os.path.isfile(repo) and os.remove(repo) is not None:
        _print("WARN: Removing repository %s failed." % repo)
        return False
    return True


def check_repo(name, check_if_enabled=True):
    """Checks if repository works and is enabled"""
    if not name:
        print("FAIL: repo name argument required")
        return False

    cmd = "yum repoinfo %s | grep Repo-status" % name  # yum=dnf alias works here
    ret, out = run(cmd, return_output=True)
    if ret != 0:
        print("%s repo is not present" % name)
        return False
    if check_if_enabled and "enabled" not in out:
        print("%s repo is not enabled" % name)
        return False

    return True


def is_docker():
    """Check if we are running inside docker container"""
    cmd = "cat /proc/self/cgroup | grep docker"
    ret = run(cmd, verbose=False, return_output=False)
    if ret == 0:
        # It is docker
        return True
    return False


def get_memory(units="m", total=False):
    """Returns data from 'free' as a dict."""
    possible_units = "b bytes k kilo m mega  g giga tera peta".split()
    if units not in possible_units:
        _print("FAIL: 'units' must be one of %s" % [str(x) for x in possible_units])
        return None

    memory = {}
    columns = []

    if len(units) > 1:
        units = "-" + units
    cmd = "free -%s" % units
    if total:
        cmd += " -t"
    ret, mem = run(cmd=cmd, return_output=True)
    if ret != 0:
        _print("FAIL: Running '%s' failed." % cmd)
        return None

    for row, m in enumerate(mem.splitlines()):
        if row == 0:
            columns = [c.strip() for c in m.split()]
            continue
        m = [x.strip() for x in m.split()]
        key = m.pop(0)[:-1].lower()
        memory[key] = {}
        for i, value in enumerate(m):
            memory[key][columns[i]] = int(value)

    return memory


def get_service_timestamp(service_name):
    """Returns active enter timestamp of a service.
    :param service_name: Name of the service
    :return:
    Time in format: a YYYY-MM-DD hh:mm:ss Z
    None: systemctl is not installed or timestamp does not exist
    """
    if not exists("systemctl"):
        cmd = "systemctl show %s --property=ActiveEnterTimestamp" % service_name
        ret, data = run(cmd, return_output=True)
        if ret == 0:
            timestamp = data.split("=")
            if timestamp[1] != "":
                return timestamp[1]
            return None
        _print("WARN: Could not get active enter timestamp of service: %s" % service_name)
    return None


def get_system_logs(
    length=None,
    reverse=False,
    kernel_only=False,
    since=None,
    grep=None,
    options=None,
    verbose=False,
    return_output=True,
):
    """Gets system logs using journalctl.
    :param length: (int) Get last $length messages.
    :param reverse: (bool) Get logs in reverse.
    :param kernel_only: (bool) Get only kernel messages.
    :param since: (string) Get messages since some time, can you '+' and '-' prefix.
    :param grep: (string) String to filter messages using 'grep'.
    :param options: (list of strings) Any other possible options with its value as a string.
    :param verbose: (bool) Print the journal when getting it.
    :param return_output (bool) Should the function return only retcode or also the output
    :return: retcode / (retcode, data)
    """
    cmd = "journalctl"
    if kernel_only:
        cmd += " -k"
    if length:
        cmd += " -n %s" % length
    if reverse:
        cmd += " -r"
    if since:
        # since can be used with '+' and '-', see man journalctl
        cmd += " -S %s" % since
    if options:
        cmd += " " + " ".join(options)

    if grep:
        cmd += " | grep '%s'" % grep

    ret, journal = run(cmd, return_output=return_output, verbose=verbose)
    if ret:
        _print(f"FAIL: cmd '{cmd}' failed with retcode {ret}.")
        return None
    if not return_output:
        return ret

    # shorten the hostname to match /var/log/messages format
    data = ""
    for line in journal.splitlines():
        line = line.split()
        if len(line) < 4:
            continue
        line[3] = line[3].split(".")[0]
        data += " ".join(line) + "\n"
    return ret, data


def edit_config(file, parameters, update=False, file_error=True, list_values=True):
    """Merges configuration file with dict using configobj
    For more complex operations see https://configobj.readthedocs.io
    :param file: (str) Path to the configuration file.
    :param parameters: (dict) Accepts also nested sections, e.g. {'logging':{'a': 'a.log', 'b': 'b.log'}}.
    :param update: (bool) Use update() instead of merge(). Will remove section entries not in parameters dict.
    :param file_error: (bool) Set False to allow creating new files.
    :param list_values: (bool) If False, values are not parsed for list values. Single line values are not unquoted.
    :return: boolean
    """
    try:
        config = ConfigObj(file, file_error=file_error, list_values=list_values)
    except OSError as e:
        print(e)
        return False

    config.filename = file

    _print(f"INFO: Updating {file}")
    if update:
        config.update(parameters)
    else:
        config.merge(parameters)

    config.write()
    return True


def remove_from_config(file, parameters_to_remove, section=None, warn=True):
    """Removes objects from a configuration file.
    :param file: (str) Path to the configuration file.
    :param parameters_to_remove: (list) Parameters to remove.
    :param section: (str) Specify section of config file parameters are nested in.
    :param warn: (bool) Set False to supress warning when deleting nonexistent parameter.
    :return: boolean
    """
    try:
        config = ConfigObj(file, file_error=True)
    except OSError as e:
        _print("FAIL: Unable to open {}. It might not exist")
        print(e)
        return False

    config.filename = file
    fail = False
    for param in parameters_to_remove:
        _print(f'INFO: Removing "{param}" from {file}')
        try:
            if section:
                config[section].pop(param)
            else:
                config.pop(param)
        except KeyError:
            if warn:
                _print("WARN: Parameter to remove is not in config")
            fail = True

    config.write()
    if fail:
        return False
    return True


def generate_sosreport(skip_plugins=None, plugin_timeout=300, timeout=900):
    """Generates a sos report.
    :param skip_plugins: (string) comma separated list of plugins to skip (no space after comma)
    :param plugin_timeout: (int) timeout in seconds to allow each plugin to run for (only applicable to rhel-8+)
    :param timeout: (int) timeout for the sosreport proces in seconds. Use `None` for no timeout.
    :return:
    """
    cmd = f"sos report --batch --plugin-timeout {plugin_timeout}"
    dist = dist_ver()
    if dist < 8:
        cmd = "sosreport --batch"

    if not install_package("sos", check=True):
        print("FAIL: unable to install sos package")
        return False

    mount_flag = False
    if is_mounted("/var/crash"):
        print("INFO: Umounting /var/crash to avoid sosreport being hang there")
        umount("/var/crash")
        mount_flag = True

    if skip_plugins:
        cmd = cmd + f" --skip-plugins {skip_plugins}"

    ret_code, sosreport_ret = run(cmd, return_output=True, timeout=timeout)
    if ret_code != 0:
        print("FAIL: sosreport command failed")
        if mount_flag:
            mount("/var/crash")
        return False

    sos_report = None
    for line in sosreport_ret.split("\n"):
        if "/tmp/sosreport" in line:
            sos_report = line.strip()
            break

    if mount_flag:
        mount("/var/crash")

    return sos_report



#!/usr/bin/python
import pytest

# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.
from libsan.host import scsi


def test_query_scsi_hosts():
    hosts = scsi.get_hosts()
    if not hosts:
        print("SKIP: Could not find scsi hosts")
        return

    for host in hosts:
        print("Querying info for host %s" % host)
        info = scsi.query_scsi_host_info(host)
        if not info:
            pytest.fail("Could not query info for host: %s" % host)
        for inf in info:
            print(f"\t{inf}: {info[inf]}")

    assert 1


def test_query_scsi_disks():
    disks = scsi.query_all_scsi_disks()
    if not disks:
        print("SKIP: Could not find scsi disks")
        return

    for disk in disks:
        print("INFO: details for scsi ID: %s" % disk)
        disk_info = disks[disk]
        for info in disk_info:
            print(f"\t{info}: {disk_info[info]}")

    assert 1



#!/usr/bin/python

# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest.mock import patch

import pytest
from libsan.host import linux, scsi_debug


def test_os_info():
    os_info = linux.query_os_info()
    if not os_info:
        pytest.fail("FAIL: Could not query OS info")

    for info in os_info:
        print(f"{info}: {os_info[info]}")
    assert 1


def test_using_scsi_debug():
    if linux.is_docker():
        print("SKIP: Cannot create scsi_debug in container")
        return
    if not scsi_debug.scsi_debug_load_module():
        pytest.fail("FAIL: loading scsi_debug module")

    device = scsi_debug.get_scsi_debug_devices()[-1]

    if linux.get_parent_device(device) != device:  # scsi_debug does not have parent
        pytest.fail("FAIL: get_parent_device using scsi_debug")
    if linux.get_full_path(device) != "/dev/%s" % device:
        pytest.fail("FAIL: get_full_path using scsi_debug")
    wwid = linux.get_device_wwid(device)
    if not wwid:
        pytest.fail("FAIL: get_device_wwid using scsi_debug")
    if linux.get_udev_property(device, "ID_MODEL") != "scsi_debug":
        pytest.fail("FAIL: device ID_MODEL should be 'scsi_debug'")
    if linux.is_dm_device(device):
        pytest.fail("FAIL: scsi_debug device should not be mapped by DM")

    if not scsi_debug.scsi_debug_unload_module():
        pytest.fail("FAIL: loading scsi_debug module")
    assert 1


def test_get_boot_device():
    get_boot = linux.get_boot_device()
    if get_boot is None and linux.is_docker() is False:
        pytest.fail("FAIL: Could not find '/boot' and '/'! ")
    print("INFO: Boot device is: %s" % get_boot)
    assert 1


def test_get_dist_release():
    release = linux.dist_release()
    if not release:
        pytest.fail("FAIL: Could not find distribution release")


def test_get_dist_ver():
    version = linux.dist_ver()
    if not version:
        pytest.fail("FAIL: Could not find distribution version")


def test_get_dist_name():
    name = linux.dist_name()
    if not name:
        pytest.fail("FAIL: Could not find distribution name")


@patch("requests.get")
def test_rpm_repo(get_mock):
    class MockResponse:
        content = b"""[test-fedora-debuginfo]
name=testrepo
#baseurl=http://download.example/pub/fedora/linux/releases/$releasever/Everything/$basearch/debug/tree/
metalink=https://mirrors.fedoraproject.org/metalink?repo=fedora-debug-$releasever&arch=$basearch
enabled=1
repo_gpgcheck=0
type=rpm
skip_if_unavailable=False
"""

    resp = MockResponse()
    get_mock.return_value = resp
    repo_metalink = "https://mirrors.fedoraproject.org/metalink?repo=fedora-debug-$releasever&arch=$basearch"
    repo_url = "https://testurl.com/test.repo"
    repo_name = "testrepo"
    linux.add_repo("test1", repo_metalink, metalink=True)
    if not linux.check_repo("test1"):
        pytest.fail("FAIL: Created repo is not working")
    if not linux.del_repo("test1"):
        pytest.fail("FAIL: Unable to delete repo")
    linux.download_repo_file(repo_url)
    if not linux.check_repo(repo_name):
        pytest.fail("FAIL: Test repo is not working")
    if not linux.del_repo("test"):
        pytest.fail("FAIL: Unable to delete repo")
    if os.path.exists("/etc/yum.repos.d/test.repo"):
        pytest.fail("FAIL: repo file should have been deleted")


@patch("libsan.host.linux.run")
def test_is_mounted(run_func):
    run_func.return_value = 0
    if not linux.is_mounted("testdev"):
        pytest.fail("FAIL: device should have been mounted")
    if not linux.is_mounted(mountpoint="testdir"):
        pytest.fail("FAIL: mountpoint should have been mounted")
    run_func.return_value = 1
    if linux.is_mounted("testdev"):
        pytest.fail("FAIL: device should NOT have been mounted")
    if linux.is_mounted(mountpoint="testdir"):
        pytest.fail("FAIL: mountpoint should NOT have been mounted")



# Copyright (C) 2018 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""qemu_img.py: Module to manipulate disk image using QEMU disk image utility."""

__author__ = "Filip Suba"
__copyright__ = "Copyright (c) 2018 Red Hat, Inc. All rights reserved."

import os

from libsan import _print
from libsan.host import linux
from libsan.host.cmdline import run

_QCOW_SUPPORTED_OPTIONS = ["compat", "backing_file", "encryption", "cluster_size", "preallocation", "lazy_refcounts"]


def _get_package_name():
    return "qemu-img"


def _get_image_file(name, image_path):
    return f"{image_path}/{name}.img"


def get_qcow_supported_options():
    """Return supported options for qcow image.
    :return: List of strings.
    """
    return _QCOW_SUPPORTED_OPTIONS


def install_qemu_img():
    """Install qemu-img tool
    :return:
    \tTrue: If qemu-img is installed correctly
    \tFalse: If some problem happened
    """
    if not linux.install_package(_get_package_name()):
        _print("FAIL: Could not install %s" % _get_package_name())
        return False
    return True


def qemu_create(filename, size="1024", fmt=None, img_path="/var/tmp", **options):
    """Create the new disk image
    :param filename: is a disk image filename
    :param size: is the disk image size in bytes
    :param fmt: is the disk image format
    :param img_path: is the full path to output directory
    :param options: see supported options for a qcow image
    :return:
    True: if success
    False: in case of failure
    """
    if not linux.is_installed(_get_package_name()):
        install_qemu_img()
    if not filename:
        _print("FAIL: qemu_create() requires parameter filename")
        return False
    if img_path:
        filename = _get_image_file(filename, img_path)
    cmd = _get_package_name() + " create "
    if fmt is not None:
        cmd += "-f %s" % fmt
    if fmt == "qcow2" and options:
        cmd += " -o "
        option = [str(i) + "=" + str(options[i]) for i in options if i in _QCOW_SUPPORTED_OPTIONS]
        cmd += ",".join(option)
    cmd += f" {filename} {size}"
    ret = run(cmd)
    if ret != 0:
        _print("FAIL: Could not create disk image.")
        return False
    return True


def delete_image(name, image_path="/var/tmp"):
    """Delete the disk image
    :param name: is the image filename
    :param image_path: is the full path to the image directory
    :return:
    True: if success
    False: in case of failure
    """
    if not name:
        _print("FAIL: delete_image() - requires name parameter")
        return False

    print("INFO: Deleting image device %s" % name)
    fname = _get_image_file(name, image_path)
    if os.path.isfile(fname):
        cmd = "rm -f %s" % fname
        retcode = run(cmd)
        if retcode != 0:
            _print("FAIL: Could not delete image disk file %s" % fname)
            return False
    return True
